import pytest


def test_verify_corruption_in_commit_ref_alerts(two_commit_filled_samples_repo):
    from hangar.records.parsing import commit_ref_db_key_from_raw_key
    from hangar.records.parsing import commit_ref_raw_val_from_db_val
    from hangar.records.parsing import commit_ref_db_val_from_raw_val

    repo = two_commit_filled_samples_repo
    history = repo.log(return_contents=True)
    head_commit = history['head']

    refKey = commit_ref_db_key_from_raw_key(head_commit)
    with repo._env.refenv.begin(write=True) as txn:
        refVal = txn.get(refKey)
        ref_unpacked = commit_ref_raw_val_from_db_val(refVal)

        modified_ref = list(ref_unpacked.db_kvs)
        modified_ref[0] = list(modified_ref[0])
        modified_ref[0][1] = b'corrupt!'
        modified_ref[0] = tuple(modified_ref[0])
        modified_ref = tuple(modified_ref)
        modifiedVal = commit_ref_db_val_from_raw_val(modified_ref)

        txn.put(refKey, modifiedVal.raw, overwrite=True)

    with pytest.raises(IOError):
        _ = repo.checkout(write=True)
    with pytest.raises(IOError):
        _ = repo.checkout(write=False)
    with pytest.raises(IOError):
        _ = repo.checkout(write=False, commit=head_commit)


def test_verify_corruption_in_commit_parent_val_alerts(two_commit_filled_samples_repo):
    from hangar.records.parsing import commit_parent_db_key_from_raw_key
    from hangar.records.parsing import commit_parent_raw_val_from_db_val
    from hangar.records.parsing import commit_parent_db_val_from_raw_val

    repo = two_commit_filled_samples_repo
    history = repo.log(return_contents=True)
    head_commit = history['head']

    parentKey = commit_parent_db_key_from_raw_key(head_commit)
    with repo._env.refenv.begin(write=True) as txn:
        parentVal = txn.get(parentKey)

        parent_raw = commit_parent_raw_val_from_db_val(parentVal)
        parent = parent_raw.ancestor_spec
        modifiedVal = commit_parent_db_val_from_raw_val(
            master_ancestor='corrupt',
            dev_ancestor=parent.dev_ancestor,
            is_merge_commit=parent.is_merge_commit)

        txn.put(parentKey, modifiedVal.raw, overwrite=True)

    with pytest.raises(IOError):
        _ = repo.checkout(write=True)
    with pytest.raises(IOError):
        _ = repo.checkout(write=False)
    with pytest.raises(IOError):
        _ = repo.checkout(write=False, commit=head_commit)


def test_verify_corruption_in_spec_val_alerts(two_commit_filled_samples_repo):
    from hangar.records.parsing import commit_spec_db_key_from_raw_key
    from hangar.records.parsing import commit_spec_db_val_from_raw_val
    from hangar.records.parsing import commit_spec_raw_val_from_db_val

    repo = two_commit_filled_samples_repo
    history = repo.log(return_contents=True)
    head_commit = history['head']

    specKey = commit_spec_db_key_from_raw_key(head_commit)
    with repo._env.refenv.begin(write=True) as txn:
        specVal = txn.get(specKey)

        spec_raw = commit_spec_raw_val_from_db_val(specVal)
        modified_spec = spec_raw.user_spec
        modified_spec = modified_spec._replace(commit_time=10.42)
        modifiedVal = commit_spec_db_val_from_raw_val(*modified_spec)

        txn.put(specKey, modifiedVal.raw, overwrite=True)

    with pytest.raises(IOError):
        _ = repo.checkout(write=True)
    with pytest.raises(IOError):
        _ = repo.checkout(write=False)
    with pytest.raises(IOError):
        _ = repo.checkout(write=False, commit=head_commit)



import pytest
import numpy as np


def test_merge_fails_with_invalid_branch_name(repo_1_br_no_conf):
    with pytest.raises(ValueError):
        cmt_hash = repo_1_br_no_conf.merge('merge commit', 'master', 'failbranchname')
    # no message passed in
    with pytest.raises(TypeError):
        cmt_hash = repo_1_br_no_conf.merge('master', 'testbranch')


def test_is_ff_merge(repo_1_br_no_conf):
    testbranch_head = repo_1_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    cmt_hash = repo_1_br_no_conf.merge('merge commit', 'master', 'testbranch')
    assert cmt_hash == testbranch_head


def test_writer_checkout_ff_merge(repo_1_br_no_conf):
    testbranch_head = repo_1_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    co = repo_1_br_no_conf.checkout(write=True, branch='master')
    master_head = co.commit_hash
    mergeHash = co.merge('dummy message', 'testbranch')
    assert mergeHash == testbranch_head
    assert mergeHash != master_head
    assert co.branch_name == 'master'
    co.close()

    master_order = repo_1_br_no_conf.log(branch='testbranch', return_contents=True)['order']
    tesbranch_order = repo_1_br_no_conf.log(branch='master', return_contents=True)['order']
    assert master_order == tesbranch_order


def test_merge_fails_if_changes_staged(repo_1_br_no_conf):
    co = repo_1_br_no_conf.checkout(write=True, branch='master')
    co.add_str_column('DOESNOTEXIST')
    co['DOESNOTEXIST'][2] = 'lol'
    co.close()
    with pytest.raises(RuntimeError, match='Changes are currently pending'):
        repo_1_br_no_conf.merge('merge commit', 'master', 'testbranch')


def test_writer_checkout_merge_fails_if_changes_staged(repo_1_br_no_conf):
    co = repo_1_br_no_conf.checkout(write=True, branch='master')
    co.add_str_column('DOESNOTEXIST')
    co['DOESNOTEXIST'][2] = 'lol'
    with pytest.raises(RuntimeError, match='Changes are currently pending'):
        co.merge('merge commit', 'testbranch')
    co.close()


def test_ff_merge_no_conf_correct_contents_for_name_or_hash_checkout(repo_1_br_no_conf):
    cmt_hash = repo_1_br_no_conf.merge('merge commit', 'master', 'testbranch')
    coByName = repo_1_br_no_conf.checkout(branch='master')
    coByHash = repo_1_br_no_conf.checkout(commit=cmt_hash)

    assert len(coByHash.columns) == len(coByName.columns)
    for asetn in coByHash.columns.keys():
        aset_byHash = coByHash.columns[asetn]
        aset_byName = coByName.columns[asetn]
        assert len(aset_byHash) == len(aset_byHash)
        for k, v in aset_byHash.items():
            assert np.allclose(v, aset_byName[k])

    coByHash.close()
    coByName.close()


def test_ff_merge_no_conf_updates_head_commit_of_branches(repo_1_br_no_conf):
    repo = repo_1_br_no_conf
    co = repo.checkout(write=True, branch='master')
    co.close()
    repo.create_branch('NotUpdatedBranch')
    old_branch_head = repo.log(branch='NotUpdatedBranch', return_contents=True)['head']

    cmt_hash = repo.merge('merge commit', 'master', 'testbranch')
    master_head = repo.log(branch='master', return_contents=True)['head']
    testbranch_head = repo.log(branch='testbranch', return_contents=True)['head']
    assert master_head == testbranch_head
    assert cmt_hash == master_head

    check_old_branch = repo.log(branch='NotUpdatedBranch', return_contents=True)['head']
    assert check_old_branch == old_branch_head
    assert check_old_branch != master_head


def test_is_3_way_merge(repo_2_br_no_conf):
    testbranch_head = repo_2_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    masterbranch_head = repo_2_br_no_conf.log(branch='master', return_contents=True)['head']
    cmt_hash = repo_2_br_no_conf.merge('merge commit', 'master', 'testbranch')
    assert cmt_hash != testbranch_head
    assert cmt_hash != masterbranch_head


def test_writer_checkout_is_3_way_merge(repo_2_br_no_conf):
    testbranch_head = repo_2_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    masterbranch_head = repo_2_br_no_conf.log(branch='master', return_contents=True)['head']
    co = repo_2_br_no_conf.checkout(write=True, branch='master')
    cmt_hash = co.merge('merge commit', 'testbranch')
    co.close()
    assert cmt_hash != testbranch_head
    assert cmt_hash != masterbranch_head


def test_3_way_merge_no_conflict_correct_contents(repo_2_br_no_conf):
    cmt_hash = repo_2_br_no_conf.merge('merge commit', 'master', 'testbranch')
    co = repo_2_br_no_conf.checkout(branch='master')
    # columns
    assert len(co.columns) == 1
    assert 'dummy' in co.columns
    # column samples
    aset = co.columns['dummy']
    assert len(aset) == 50

    # column sample values
    checkarr = np.zeros_like(np.arange(50))
    for k, v in aset.items():
        checkarr[:] = int(k)
        assert np.allclose(v, checkarr)

    # column sample keys
    aset_keys = list(aset.keys())
    for genKey in range(30):
        assert str(genKey) in aset_keys
        aset_keys.remove(str(genKey))
    assert len(aset_keys) == 20
    co.close()


def test_writer_checkout_3_way_merge_no_conflict_correct_contents(repo_2_br_no_conf):
    co = repo_2_br_no_conf.checkout(write=True, branch='master')
    cmt_hash = co.merge('merge commit', 'testbranch')

    # columns
    assert len(co.columns) == 1
    assert 'dummy' in co.columns
    # column samples
    aset = co.columns['dummy']
    assert len(aset) == 50

    # column sample values
    checkarr = np.zeros_like(np.arange(50))
    for k, v in aset.items():
        checkarr[:] = int(k)
        assert np.allclose(v, checkarr)

    # column sample keys
    aset_keys = list(aset.keys())
    for genKey in range(30):
        assert str(genKey) in aset_keys
        aset_keys.remove(str(genKey))
    assert len(aset_keys) == 20
    co.close()


def test_3_way_merge_no_conflict_and_mutation_correct_contents(repo_2_br_no_conf):
    co = repo_2_br_no_conf.checkout(write=True, branch='master')
    co.columns['dummy']['1'] = co.columns['dummy']['0']
    co.commit('mutated master')
    co.close()

    co = repo_2_br_no_conf.checkout(write=True, branch='testbranch')
    co.columns['dummy']['2'] = co.columns['dummy']['0']
    co.commit('mutated testbranch')
    co.close()

    repo_2_br_no_conf.merge('merge commit', 'master', 'testbranch')
    co = repo_2_br_no_conf.checkout(branch='master')

    # columns
    assert len(co.columns) == 1
    assert 'dummy' in co.columns
    # column samples
    aset = co.columns['dummy']
    assert len(aset) == 50

    # column sample values
    checkarr = np.zeros_like(np.arange(50))
    for k, v in aset.items():
        if k == '2':
            checkarr[:] = 0
        elif k == '1':
            checkarr[:] = 0
        else:
            checkarr[:] = int(k)
        assert np.allclose(v, checkarr)

    # column sample keys
    aset_keys = list(aset.keys())
    for genKey in range(30):
        assert str(genKey) in aset_keys
        aset_keys.remove(str(genKey))
    assert len(aset_keys) == 20
    co.close()


def test_3_way_merge_updates_head_commit_of_branches(repo_2_br_no_conf):
    orig_testbranch_head = repo_2_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    orig_masterbranch_head = repo_2_br_no_conf.log(branch='master', return_contents=True)['head']

    cmt_hash = repo_2_br_no_conf.merge('merge commit', 'master', 'testbranch')

    new_testbranch_head = repo_2_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    new_masterbranch_head = repo_2_br_no_conf.log(branch='master', return_contents=True)['head']

    assert orig_testbranch_head == new_testbranch_head
    assert orig_masterbranch_head != new_masterbranch_head
    assert new_masterbranch_head == cmt_hash


def test_writer_checkout_3_way_merge_updates_head_commit_of_branches(repo_2_br_no_conf):
    orig_testbranch_head = repo_2_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    orig_masterbranch_head = repo_2_br_no_conf.log(branch='master', return_contents=True)['head']

    co = repo_2_br_no_conf.checkout(write=True, branch='master')
    cmt_hash = co.merge('merge commit', 'testbranch')
    assert cmt_hash == co.commit_hash
    co.close()

    new_testbranch_head = repo_2_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    new_masterbranch_head = repo_2_br_no_conf.log(branch='master', return_contents=True)['head']
    assert orig_testbranch_head == new_testbranch_head
    assert orig_masterbranch_head != new_masterbranch_head
    assert new_masterbranch_head == cmt_hash


class TestArraysetSampleConflicts(object):

    def test_conflict_additions_same_str_name_different_value(self, repo_2_br_no_conf):
        newdata = np.arange(50)
        newdata = newdata * 2

        repo = repo_2_br_no_conf
        co = repo.checkout(write=True, branch='master')
        co.columns['dummy']['15'] = newdata
        co.commit('commit on master with conflicting data')
        co.close()

        with pytest.raises(ValueError):
            repo.merge('merge commit', 'master', 'testbranch')

    def test_conflict_additions_same_int_name_different_value(self, repo_2_br_no_conf):
        newdata = np.arange(50)
        newdata = newdata * 2

        repo = repo_2_br_no_conf
        co = repo.checkout(write=True, branch='master')
        co.columns['dummy'][15] = newdata
        co.commit('commit on master with conflicting data')
        co.close()

        with pytest.raises(ValueError):
            repo.merge('merge commit', 'master', 'testbranch')

    def test_conflict_additions_same_str_and_int_name_different_value(self, repo_2_br_no_conf):
        newdata = np.arange(50)
        newdata = newdata * 2

        repo = repo_2_br_no_conf
        co = repo.checkout(write=True, branch='master')
        co.columns['dummy'][15] = newdata
        co.columns['dummy']['15'] = newdata
        co.commit('commit on master with conflicting data')
        co.close()

        with pytest.raises(ValueError):
            repo.merge('merge commit', 'master', 'testbranch')

    def test_no_conflict_additions_same_name_and_value(self, repo_2_br_no_conf):
        newdata = np.arange(50)
        newdata[:] = 15

        repo = repo_2_br_no_conf
        co = repo.checkout(write=True, branch='master')
        co.columns['dummy']['15'] = newdata
        co.columns['dummy'][15] = newdata
        co.commit('commit on master with same value data')
        co.close()

        cmt_hash = repo.merge('merge commit', 'master', 'testbranch')
        co = repo.checkout(commit=cmt_hash)
        aset = co.columns['dummy']
        assert np.allclose(aset['15'], newdata)
        assert np.allclose(aset[15], newdata)
        co.close()

    def test_conflict_mutations_same_name_different_value(self, repo_2_br_no_conf):
        repo = repo_2_br_no_conf
        co = repo.checkout(write=True, branch='master')
        newdata = np.arange(50)
        co.columns['dummy']['0'] = newdata
        co.commit('commit on master with conflicting data')
        co.close()

        co = repo.checkout(write=True, branch='testbranch')
        newdata = newdata * 2
        co.columns['dummy']['0'] = newdata
        co.commit('commit on testbranch with conflicting data')
        co.close()

        with pytest.raises(ValueError):
            repo.merge('merge commit', 'master', 'testbranch')

    def test_conflict_mutation_and_removal(self, repo_2_br_no_conf):
        repo = repo_2_br_no_conf
        co = repo.checkout(write=True, branch='master')
        newdata = np.arange(50)
        co.columns['dummy']['0'] = newdata
        co.commit('commit on master with conflicting data')
        co.close()

        co = repo.checkout(write=True, branch='testbranch')
        del co.columns['dummy']['0']
        co.commit('commit on testbranch with removal')
        co.close()

        with pytest.raises(ValueError):
            repo.merge('merge commit', 'master', 'testbranch')

    def test_no_conflict_both_removal(self, repo_2_br_no_conf):
        repo = repo_2_br_no_conf
        co = repo.checkout(write=True, branch='master')
        del co.columns['dummy']['0']
        del co.columns['dummy'][21]
        co.commit('commit on master with removal')
        co.close()

        co = repo.checkout(write=True, branch='testbranch')
        del co.columns['dummy']['0']
        del co.columns['dummy'][10]
        co.commit('commit on testbranch with removal')
        co.close()

        cmt_hash = repo.merge('merge commit', 'master', 'testbranch')
        co = repo.checkout(commit=cmt_hash)
        aset = co.columns['dummy']
        assert '0' not in aset
        assert len(aset) == 47



__all__ = ('make_torch_dataset', 'make_tf_dataset')


# lazy loader
def __getattr__(name):
    global make_tf_dataset, make_torch_dataset

    if name == 'make_tf_dataset':
        from .tfloader import make_tf_dataset as tf_dset
        make_tf_dataset = tf_dset
        return tf_dset

    if name == 'make_torch_dataset':
        from .torchloader import make_torch_dataset as torch_dset
        make_torch_dataset = torch_dset
        return torch_dset

    raise AttributeError(f"module {__name__} has no attribute {name}")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.intersphinx',
    'sphinx_click.ext',
    'nbsphinx',
    'sphinx.ext.mathjax',
    'recommonmark',
    'IPython.sphinxext.ipython_console_highlighting',
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

nbsphinx_execute = 'never'

autodoc_mock_imports = ['torch', 'tensorflow']
autosummary_generate = True

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}
master_doc = 'index'
project = 'Hangar'
year = '2019-2020'
author = 'Richard Izzo'
copyright = '{0}, {1}'.format(year, author)
version = release = '0.5.0'

pygments_style = 'default'
pygments_lexer = 'PythonConsoleLexer'
highlight_options = {
    'python3': True
}
templates_path = ['.']
exclude_patterns = ['_build', '**.ipynb_checkpoints']
extlinks = {
    'issue': ('https://github.com/tensorwerk/hangar-py/issues/%s', '#'),
    'pr': ('https://github.com/tensorwerk/hangar-py/pull/%s', 'PR #'),
}
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'torch': ('https://pytorch.org/docs/master', None),
    'numpy': ('http://docs.scipy.org/doc/numpy', None),
}

# Regular expressions that match URIs that should not be checked
# when doing a linkcheck build
linkcheck_ignore = [
    r'http://localhost:\d+/?', 'http://localhost/',
    'https://github.com/tensorwerk/hangar-py',
    r'https://github.com/tensorwerk/hangar-py/.*',
]
linkcheck_retries = 3

# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# if not on_rtd:  # only set the theme if we're building docs locally
html_theme = 'sphinx_rtd_theme'

html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = True
napoleon_use_param = True
napoleon_include_init_with_doc = True

add_module_names = False
doctest_test_doctest_blocks = None
autoclass_content = 'class'

html_theme_options = {
    'style_nav_header_background': 'orange'
}



import pytest
import numpy as np


# -------------------------- Reader Checkout ----------------------------------


@pytest.mark.parametrize('write', [True, False])
def test_arrayset_getattr_does_not_raise_permission_error_if_alive(write, aset_samples_initialized_repo):
    co = aset_samples_initialized_repo.checkout(write=write)
    asets = co.columns

    assert hasattr(asets, 'doesnotexist') is False  # does not raise error
    assert hasattr(asets, '_mode') is True
    with pytest.raises(AttributeError):
        assert getattr(asets, 'doesnotexist')
    assert getattr(asets, '_mode') == 'a' if write else 'r'

    co.close()
    with pytest.raises(PermissionError):
        hasattr(asets, 'doesnotexist')
    with pytest.raises(PermissionError):
        hasattr(asets, '_mode')


def test_write_in_context_manager_no_loop(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    with wco:
        assert wco._is_conman is True
        wco['writtenaset']['0'] = array5by7
        wco['newaset']['0'] = array10
    assert wco._is_conman is False

    assert np.allclose(array5by7, wco.columns['writtenaset']['0'])
    assert np.allclose(array10, wco.columns['newaset']['0'])
    wco.commit('init')
    assert np.allclose(array5by7, wco.columns['writtenaset']['0'])
    assert np.allclose(array10, wco.columns['newaset']['0'])
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    assert np.allclose(array5by7, rco.columns['writtenaset']['0'])
    assert np.allclose(array10, rco.columns['newaset']['0'])
    rco.close()


def test_write_in_context_manager_many_samples_looping(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    with wco:
        assert wco._is_conman is True
        for idx in range(100):
            array10[:] = idx
            array5by7[:] = idx
            wco['writtenaset'][idx] = array5by7
            wco['newaset'][idx] = array10
    assert wco._is_conman is False

    for idx in range(100):
        array10[:] = idx
        array5by7[:] = idx
        assert np.allclose(array5by7, wco.columns['writtenaset'][idx])
        assert np.allclose(array10, wco.columns['newaset'][idx])
    wco.commit('init')
    for idx in range(100):
        array10[:] = idx
        array5by7[:] = idx
        assert np.allclose(array5by7, wco.columns['writtenaset'][idx])
        assert np.allclose(array10, wco.columns['newaset'][idx])
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    for idx in range(100):
        array10[:] = idx
        array5by7[:] = idx
        assert np.allclose(array5by7, rco.columns['writtenaset'][idx])
        assert np.allclose(array10, rco.columns['newaset'][idx])
    rco.close()


def test_write_fails_if_checkout_closed(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)
    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    wco['writtenaset'][0] = array5by7
    wco['newaset'][0] = array10
    wco.close()
    with pytest.raises((PermissionError, UnboundLocalError)):
        wco['writtenaset'][1] = array5by7
        wco['newaset'][1] = array10

    wco2 = aset_samples_initialized_repo.checkout(write=True)
    assert 0 in wco2.columns['writtenaset']
    assert 0 in wco2.columns['newaset']
    assert 1 not in wco2.columns['writtenaset']
    assert 1 not in wco2.columns['newaset']
    wco2.close()


def test_write_context_manager_fails_if_checkout_closed(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)
    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    wco['writtenaset'][0] = array5by7
    wco['newaset'][0] = array10
    wco.close()
    with pytest.raises(PermissionError):
        with wco:
            wco['writtenaset'][1] = array5by7
    with pytest.raises(PermissionError):
        with wco:
            wco['newaset'][1] = array10

    wco2 = aset_samples_initialized_repo.checkout(write=True)
    assert 0 in wco2.columns['writtenaset']
    assert 0 in wco2.columns['newaset']
    assert 1 not in wco2.columns['writtenaset']
    assert 1 not in wco2.columns['newaset']
    wco2.close()


def test_writer_co_read_single_aset_single_sample(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2

    assert np.allclose(wco['writtenaset', 0], array5by7)
    assert np.allclose(wco['writtenaset', 1], array5by7 + 1)
    assert np.allclose(wco['writtenaset', 2], array5by7 + 2)
    wco.close()


def test_writer_co_read_single_aset_multiple_samples(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2

    res = wco[('writtenaset', 0), ('writtenaset', 1), ('writtenaset', 2)]
    assert np.allclose(res[0], array5by7)
    assert np.allclose(res[1], array5by7 + 1)
    assert np.allclose(res[2], array5by7 + 2)
    wco.close()


def test_writer_co_read_multiple_aset_single_samples(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    array10[:] = 0
    wco.columns['newaset'][0] = array10
    wco.columns['newaset'][1] = array10 + 1
    wco.columns['newaset'][2] = array10 + 2

    res = wco[('writtenaset', 0), ('newaset', 0)]
    assert np.allclose(res[0], array5by7)
    assert np.allclose(res[1], array10)
    res = wco[('writtenaset', 1), ('newaset', 1)]
    assert np.allclose(res[0], array5by7 + 1)
    assert np.allclose(res[1], array10 + 1)
    wco.close()


def test_writer_co_read_multtiple_aset_multiple_samples(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    array10[:] = 0
    wco.columns['newaset'][0] = array10
    wco.columns['newaset'][1] = array10 + 1
    wco.columns['newaset'][2] = array10 + 2

    res = wco[('writtenaset', 0), ('newaset', 0), ('writtenaset', 1), ('newaset', 1)]
    assert isinstance(res, list)
    assert len(res) == 4
    assert np.allclose(res[0], array5by7)
    assert np.allclose(res[1], array10)
    assert np.allclose(res[2], array5by7 + 1)
    assert np.allclose(res[3], array10 + 1)
    wco.close()


def test_writer_co_read_fails_nonexistant_aset_name(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    with pytest.raises(KeyError):
        _ = wco['doesnotexist', 0]
    wco.close()


def test_writer_co_read_fails_nonexistant_sample_name(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    with pytest.raises(KeyError):
        _ = wco['doesnotexist', 124]
    wco.close()


def test_writer_co_get_returns_none_on_nonexistant_sample_name(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    out = wco.get(('writtenaset', 124))
    assert out is None
    wco.close()


def test_writer_co_read_in_context_manager_no_loop(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    wco['writtenaset']['0'] = array5by7
    wco['newaset']['0'] = array10
    with wco:
        assert wco._is_conman is True
        assert np.allclose(wco['writtenaset', '0'], array5by7)
    wco.close()


def test_writer_co_read_in_context_manager_many_samples_looping(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    with wco:
        for idx in range(100):
            array10[:] = idx
            array5by7[:] = idx
            wco['writtenaset'][idx] = array5by7
            wco['newaset'][idx] = array10

    with wco:
        waset_keys = [('writtenaset', i) for i in range(100)]
        naset_keys = [('newaset', i) for i in range(100)]
        writtenasetOut = wco[waset_keys]
        newasetOut = wco[naset_keys]
        for idx in range(100):
            array10[:] = idx
            array5by7[:] = idx
            assert np.allclose(array5by7, wco['writtenaset', idx])
            assert np.allclose(array10, wco['newaset', idx])

            o = wco[('writtenaset', idx), ('newaset', idx)]
            assert np.allclose(o[0], array5by7)
            assert np.allclose(o[1], array10)

            assert np.allclose(writtenasetOut[idx], array5by7)
            assert np.allclose(newasetOut[idx], array10)
    wco.close()


@pytest.mark.parametrize('write', [True, False])
def test_co_read_dunder_getitem_excepts_missing_sample(aset_samples_initialized_repo, write):
    co = aset_samples_initialized_repo.checkout(write=write)
    with pytest.raises(KeyError):
        res = co['writtenaset', 0]
    co.close()


@pytest.mark.parametrize('write', [True, False])
def test_co_read_get_except_missing_true_excepts_missing_sample(aset_samples_initialized_repo, write):
    co = aset_samples_initialized_repo.checkout(write=write)
    with pytest.raises(KeyError):
        res = co.get(('writtenaset', 0), except_missing=True)
    co.close()


@pytest.mark.parametrize('write', [True, False])
def test_co_read_get_except_missing_false_returns_none_on_missing_sample(aset_samples_initialized_repo, write):
    co = aset_samples_initialized_repo.checkout(write=write)
    res_1 = co.get(('writtenaset', 0))
    assert res_1 is None
    res_2 = co.get(('writtenaset', 0), except_missing=False)
    assert res_2 is None
    co.close()


def test_writer_co_aset_finds_connection_manager_of_any_aset_in_cm(aset_samples_initialized_repo):
    wco = aset_samples_initialized_repo.checkout(write=True)
    wco.add_ndarray_column('second', shape=(20,), dtype=np.uint8)
    asets = wco.columns

    with wco.columns['second'] as second_aset:
        assert wco.columns['second']._is_conman is True
        assert second_aset._is_conman is True
        assert asets._any_is_conman() is True

    with wco.columns['writtenaset'] as written_aset:
        assert wco.columns['writtenaset']._is_conman is True
        assert written_aset._is_conman is True
        assert asets._any_is_conman() is True

    assert wco.columns['writtenaset']._is_conman is False
    assert wco.columns['second']._is_conman is False
    assert asets._any_is_conman() is False
    wco.close()


def test_writer_co_aset_cm_not_allow_remove_aset(aset_samples_initialized_repo, array5by7):

    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2

    asets = wco.columns
    with asets as cm_asets:
        with pytest.raises(PermissionError):
            cm_asets.delete('writtenaset')
        with pytest.raises(PermissionError):
            asets.delete('writtenaset')
        with pytest.raises(PermissionError):
            wco.columns.delete('writtenaset')

        with pytest.raises(PermissionError):
            del cm_asets['writtenaset']
        with pytest.raises(PermissionError):
            del asets['writtenaset']
        with pytest.raises(PermissionError):
            del wco.columns['writtenaset']

    assert len(wco['writtenaset']) == 3
    assert np.allclose(wco['writtenaset', 0], array5by7)
    assert np.allclose(wco['writtenaset', 1], array5by7 + 1)
    assert np.allclose(wco['writtenaset', 2], array5by7 + 2)
    wco.close()


def test_writer_co_column_instance_cm_not_allow_any_column_removal(repo_20_filled_samples):

    wco = repo_20_filled_samples.checkout(write=True)
    columns = wco.columns
    writtenaset = wco.columns['writtenaset']
    second_aset = wco.columns['second_aset']

    with second_aset:
        with pytest.raises(PermissionError):
            columns.delete('writtenaset')
        with pytest.raises(PermissionError):
            columns.delete('second_aset')
        with pytest.raises(PermissionError):
            wco.columns.delete('writtenaset')
        with pytest.raises(PermissionError):
            wco.columns.delete('second_aset')
        with pytest.raises(PermissionError):
            del columns['writtenaset']
        with pytest.raises(PermissionError):
            del columns['second_aset']
        with pytest.raises(PermissionError):
            del wco.columns['second_aset']
        with pytest.raises(PermissionError):
            del wco.columns['written_aset']

    with writtenaset:
        with pytest.raises(PermissionError):
            columns.delete('writtenaset')
        with pytest.raises(PermissionError):
            columns.delete('second_aset')
        with pytest.raises(PermissionError):
            wco.columns.delete('writtenaset')
        with pytest.raises(PermissionError):
            wco.columns.delete('second_aset')
        with pytest.raises(PermissionError):
            del columns['writtenaset']
        with pytest.raises(PermissionError):
            del columns['second_aset']
        with pytest.raises(PermissionError):
            del wco.columns['second_aset']
        with pytest.raises(PermissionError):
            del wco.columns['written_aset']

    with columns:
        with pytest.raises(PermissionError):
            columns.delete('writtenaset')
        with pytest.raises(PermissionError):
            columns.delete('second_aset')
        with pytest.raises(PermissionError):
            wco.columns.delete('writtenaset')
        with pytest.raises(PermissionError):
            wco.columns.delete('second_aset')
        with pytest.raises(PermissionError):
            del columns['writtenaset']
        with pytest.raises(PermissionError):
            del columns['second_aset']
        with pytest.raises(PermissionError):
            del wco.columns['second_aset']
        with pytest.raises(PermissionError):
            del wco.columns['written_aset']

    wco.close()


def test_writer_co_aset_removes_all_samples_and_arrayset_still_exists(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)
    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2
    assert len(wco.columns) == 1
    assert len(wco.columns['writtenaset']) == 3

    with wco.columns['writtenaset'] as wset:
        del wset[0]
        del wset[1]
        del wset[2]
        # Removed all samples, now the aset's gone
        assert len(wset) == 0
        assert len(wco.columns) == 1
    assert len(wco.columns) == 1

    del wco.columns['writtenaset']

    assert len(wco.columns) == 0
    with pytest.raises(KeyError):
        len(wco.columns['writtenaset'])
    wco.close()


# -------------------------- Reader Checkout ----------------------------------


def test_reader_co_read_single_aset_single_sample(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2
    wco.commit('first')
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    assert np.allclose(rco['writtenaset', 0], array5by7)
    assert np.allclose(rco['writtenaset', 1], array5by7 + 1)
    assert np.allclose(rco['writtenaset', 2], array5by7 + 2)
    rco.close()


def test_reader_co_read_single_aset_multiple_samples(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2
    wco.commit('first')
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    res = rco[('writtenaset', 0), ('writtenaset', 1), ('writtenaset', 2)]
    assert np.allclose(res[0], array5by7)
    assert np.allclose(res[1], array5by7 + 1)
    assert np.allclose(res[2], array5by7 + 2)
    rco.close()


def test_reader_co_read_multiple_aset_single_samples(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    array10[:] = 0
    wco.columns['newaset'][0] = array10
    wco.columns['newaset'][1] = array10 + 1
    wco.columns['newaset'][2] = array10 + 2
    wco.commit('first')
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    res = rco[('writtenaset', 0), ('newaset', 0)]
    assert np.allclose(res[0], array5by7)
    assert np.allclose(res[1], array10)
    res = rco[('writtenaset', 1), ('newaset', 1)]
    assert np.allclose(res[0], array5by7 + 1)
    assert np.allclose(res[1], array10 + 1)
    rco.close()


def test_reader_co_read_multtiple_aset_multiple_samples(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.columns['writtenaset'][1] = array5by7 + 1
    wco.columns['writtenaset'][2] = array5by7 + 2

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    array10[:] = 0
    wco.columns['newaset'][0] = array10
    wco.columns['newaset'][1] = array10 + 1
    wco.columns['newaset'][2] = array10 + 2
    wco.commit('first')
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    res = rco[('writtenaset', 0), ('newaset', 0), ('writtenaset', 1), ('newaset', 1)]
    assert isinstance(res, list)
    assert len(res) == 4
    assert np.allclose(res[0], array5by7)
    assert np.allclose(res[1], array10)
    assert np.allclose(res[2], array5by7 + 1)
    assert np.allclose(res[3], array10 + 1)
    rco.close()


def test_reader_co_read_fails_nonexistant_aset_name(aset_samples_initialized_repo, array5by7):
    rco = aset_samples_initialized_repo.checkout()
    with pytest.raises(KeyError):
        _ = rco['doesnotexist', 0]
    rco.close()


def test_reader_co_read_fails_nonexistant_sample_name(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)
    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.commit('first')
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    with pytest.raises(KeyError):
        _ = rco['doesnotexist', 124]
    rco.close()


def test_reader_co_get_read_returns_none_nonexistant_sample_name(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)
    array5by7[:] = 0
    wco.columns['writtenaset'][0] = array5by7
    wco.commit('first')
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    out = rco.get(('writtenaset', 124))
    assert out is None
    rco.close()


def test_reader_co_read_in_context_manager_no_loop(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    wco['writtenaset']['0'] = array5by7
    wco['newaset']['0'] = array10
    wco.commit('first')
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    with rco:
        assert rco._is_conman is True
        assert np.allclose(rco['writtenaset', '0'], array5by7)
    rco.close()


def test_reader_co_read_in_context_manager_many_samples_looping(aset_samples_initialized_repo, array5by7):
    wco = aset_samples_initialized_repo.checkout(write=True)

    array10 = np.arange(10, dtype=np.float32)
    wco.add_ndarray_column('newaset', prototype=array10)
    with wco:
        for idx in range(100):
            array10[:] = idx
            array5by7[:] = idx
            wco['writtenaset'][idx] = array5by7
            wco['newaset'][idx] = array10
    wco.commit('first')
    wco.close()

    rco = aset_samples_initialized_repo.checkout()
    with rco:
        waset_keys = [('writtenaset', i) for i in range(100)]
        naset_keys = [('newaset', i) for i in range(100)]
        writtenasetOut = rco[waset_keys]
        newasetOut = rco[naset_keys]
        for idx in range(100):
            array10[:] = idx
            array5by7[:] = idx
            assert np.allclose(array5by7, rco['writtenaset', idx])
            assert np.allclose(array10, rco['newaset', idx])

            o = rco[('writtenaset', idx), ('newaset', idx)]
            assert np.allclose(o[0], array5by7)
            assert np.allclose(o[1], array10)
            assert np.allclose(writtenasetOut[idx], array5by7)
            assert np.allclose(newasetOut[idx], array10)
    rco.close()



"""Tests for the class methods contained in the nested subsample column accessor.
"""
import pytest
import numpy as np
from conftest import fixed_shape_backend_params, variable_shape_backend_params


# --------------------------- Setup ------------------------------


def assert_equal(arr, arr2):
    assert np.array_equal(arr, arr2)
    assert arr.dtype == arr2.dtype


# ------------------------ Tests ----------------------------------


class TestArraysetSetup:

    @pytest.mark.parametrize('name', [
        'invalid\n', '\ninvalid', 'inv name', 'inva@lid', 12, ' try', 'andthis ',
        'VeryLongNameIsInvalidOver64CharactersNotAllowedVeryLongNameIsInva'])
    def test_does_not_allow_invalid_arrayset_names(self, repo, randomsizedarray, name):
        co = repo.checkout(write=True)
        with pytest.raises(ValueError):
            co.add_ndarray_column(name, prototype=randomsizedarray, contains_subsamples=True)
        co.close()

    def test_read_only_mode_arrayset_methods_limited(self, aset_subsamples_initialized_repo):
        import hangar
        co = aset_subsamples_initialized_repo.checkout()
        assert isinstance(co, hangar.checkout.ReaderCheckout)
        with pytest.raises(AttributeError):
            assert co.add_ndarray_column('foo')
        with pytest.raises(AttributeError):
            assert co.add_str_column('foo')
        with pytest.raises(PermissionError):
            assert co.columns.delete('foo')
        assert len(co.columns['writtenaset']) == 0
        co.close()

    def test_get_arrayset_in_read_and_write_checkouts(self, aset_subsamples_initialized_repo, array5by7):
        co = aset_subsamples_initialized_repo.checkout(write=True)
        # getting the column with `get`
        asetOld = co.columns.get('writtenaset')
        asetOldPath = asetOld._path
        asetOldAsetn = asetOld.column
        asetOldDefaultSchemaHash = asetOld._schema.schema_hash_digest()
        co.close()

        co = aset_subsamples_initialized_repo.checkout()
        # getting column with dictionary like style method
        asetNew = co.columns['writtenaset']
        assert asetOldPath == asetNew._path
        assert asetOldAsetn == asetNew.column
        assert asetOldDefaultSchemaHash == asetNew._schema.schema_hash_digest()
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_delete_arrayset(self, aset_backend, aset_subsamples_initialized_repo):
        co = aset_subsamples_initialized_repo.checkout(write=True)
        co.columns.delete('writtenaset')
        assert 'writtenaset' not in co.columns
        with pytest.raises(KeyError):
            # cannot delete twice
            co.columns.delete('writtenaset')

        # init and immediate delete leaves no trace
        co.add_ndarray_column(name='writtenaset', shape=(5, 7), dtype=np.float64,
                                 backend=aset_backend, contains_subsamples=True)
        assert len(co.columns) == 1
        co.columns.delete('writtenaset')
        assert len(co.columns) == 0
        co.commit('this is a commit message')
        co.close()

        # init column in checkout persists aset records/accessor even if no samples contained
        co = aset_subsamples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 0
        co.add_ndarray_column(name='writtenaset', shape=(5, 7), dtype=np.float64,
                                 backend=aset_backend, contains_subsamples=True)
        co.commit('this is a commit message')
        co.close()
        co = aset_subsamples_initialized_repo.checkout(write=True)
        assert len(co.columns) == 1

        # column can be deleted with via __delitem__ dict style command.
        del co.columns['writtenaset']
        assert len(co.columns) == 0
        co.commit('this is a commit message')
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_init_same_arrayset_twice_fails_again(self, aset_backend, repo, randomsizedarray):
        co = repo.checkout(write=True)
        co.add_ndarray_column('aset', prototype=randomsizedarray,
                                 backend=aset_backend, contains_subsamples=True)
        with pytest.raises(LookupError):
            # test if everything is the same as initalized one.
            co.add_ndarray_column('aset', prototype=randomsizedarray,
                                     backend=aset_backend, contains_subsamples=True)
        with pytest.raises(LookupError):
            # test if column container type is different than existing name (no subsamples0
            co.add_ndarray_column('aset', prototype=randomsizedarray,
                                     backend=aset_backend, contains_subsamples=False)
        co.close()

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_arrayset_with_invalid_dimension_sizes_shapes(self, aset_backend, repo):
        co = repo.checkout(write=True)

        shape = (0, 1, 2)
        with pytest.raises(ValueError):
            # cannot have zero valued size for any dimension
            co.add_ndarray_column('aset', shape=shape, dtype=np.int,
                                     backend=aset_backend, contains_subsamples=True)

        shape = [1] * 31
        aset = co.add_ndarray_column('aset1', shape=shape, dtype=np.int,
                                        backend=aset_backend, contains_subsamples=True)
        assert len(aset.shape) == 31

        shape = [1] * 32
        with pytest.raises(ValueError):
            # maximum tensor rank must be <= 31
            co.add_ndarray_column('aset2', shape=shape, dtype=np.int,
                                     backend=aset_backend, contains_subsamples=True)
        co.close()


# ------------------------------ Add Data Tests --------------------------------------------

@pytest.fixture(params=[1, 3], scope='class')
def multi_item_generator(request):
    yield request.param


@pytest.fixture(params=[
    # specifies container types, two-elements: ['outer', 'inner']
    ['dict', None],
    ['list', 'tuple'],
    ['tuple', 'list'],
], scope='class')
def iterable_subsamples(request, multi_item_generator):
    outer, inner = request.param
    arrays = []
    for num_item in range(multi_item_generator):
        arr = np.arange(16, dtype=np.uint8).reshape(4, 4)
        arr += 1
        arrays.append(arr)

    components = []
    for idx, array in enumerate(arrays):
        if inner == 'list':
            component = [f'subsample{idx}', array]
        elif inner == 'tuple':
            component = (f'subsample{idx}', array)
        elif inner is None:
            component = {f'subsample{idx}': array}
        else:
            raise ValueError(
                f'unknown parameter of `inner` {inner} in test suite generation')
        components.append(component)

    if outer == 'dict':
        res = {}
        for part in components:
            res.update(part)
    elif outer == 'list':
        res = []
        for part in components:
            res.append(part)
    elif outer == 'tuple':
        res = []
        for part in components:
            res.append(part)
        res = tuple(res)
    else:
        raise ValueError(
            f'unknown parameter of `outer` {outer} in test suite generation')
    return res


@pytest.fixture(params=['dict', 'list', 'tuple'], scope='class')
def iterable_samples(request, multi_item_generator, iterable_subsamples):
    container = request.param

    if container == 'dict':
        res = {}
        for idx in range(multi_item_generator):
            res[f'sample{idx}'] = iterable_subsamples
    elif container == 'list':
        res = []
        for idx in range(multi_item_generator):
            res.append([f'sample{idx}', iterable_subsamples])
    elif container == 'tuple':
        res = []
        for idx in range(multi_item_generator):
            res.append([f'sample{idx}', iterable_subsamples])
        res = tuple(res)
    else:
        raise ValueError(
            f'unknown parameter of `container` {container} in test suite generation')
    return res


@pytest.fixture(params=fixed_shape_backend_params, scope='class')
def backend_params(request):
    return request.param


@pytest.fixture()
def subsample_writer_written_aset(backend_params, repo, monkeypatch):
    from hangar.backends import hdf5_00
    from hangar.backends import hdf5_01
    from hangar.backends import numpy_10
    monkeypatch.setattr(hdf5_00, 'COLLECTION_COUNT', 5)
    monkeypatch.setattr(hdf5_00, 'COLLECTION_SIZE', 10)
    monkeypatch.setattr(hdf5_01, 'COLLECTION_COUNT', 5)
    monkeypatch.setattr(hdf5_01, 'COLLECTION_SIZE', 10)
    monkeypatch.setattr(numpy_10, 'COLLECTION_SIZE', 10)

    co = repo.checkout(write=True)
    aset = co.add_ndarray_column('foo', shape=(4, 4), dtype=np.uint8, variable_shape=False,
                                    backend=backend_params, contains_subsamples=True)
    yield aset
    co.close()


class TestAddData:

    def test_update_sample_subsamples_empty_arrayset(self, subsample_writer_written_aset, iterable_samples):
        aset = subsample_writer_written_aset
        added = aset.update(iterable_samples)
        assert added is None
        assert len(aset._samples) == len(iterable_samples)
        for sample_idx, sample_data in enumerate(iterable_samples):
            assert f'sample{sample_idx}' in aset._samples

    def test_update_sample_kwargs_only_empty_arrayset(self, subsample_writer_written_aset, iterable_subsamples):
        aset = subsample_writer_written_aset
        added = aset.update(fookwarg=iterable_subsamples)
        assert added is None
        assert len(aset._samples) == 1
        assert 'fookwarg' in aset._samples

        added = aset.update(bar=iterable_subsamples, baz=iterable_subsamples)
        assert added is None
        assert len(aset._samples) == 3
        assert 'bar' in aset._samples
        assert 'baz' in aset._samples
        for subsample_idx, _data in enumerate(iterable_subsamples):
            assert f'subsample{subsample_idx}' in aset._samples['fookwarg']._subsamples
            assert f'subsample{subsample_idx}' in aset._samples['bar']._subsamples
            assert f'subsample{subsample_idx}' in aset._samples['baz']._subsamples

    def test_update_sample_kwargs_and_other_dict_doesnt_modify_input_in_calling_scope(
        self, subsample_writer_written_aset, iterable_subsamples, iterable_samples
    ):
        """ensure bug does not revert.

        Had a case where if dict was passed as ``other`` along with kwargs, the operation
        would complete as normally, but when control returned to the caller the original
        dict passed in as ``other`` would have been silently merged with the kwargs.
        """
        aset = subsample_writer_written_aset
        if not isinstance(iterable_samples, dict):
            return
        iterable_samples_before = list(iterable_samples.items())

        aset.update(iterable_samples, kwargadded=iterable_subsamples)
        # in bug case, would now observe that iterable_samples would have been
        # silently modified in a method analogous to calling:
        #
        #   ``iterable_samples.update({'kwargadded': iterable_subsamples})``
        #
        assert list(iterable_samples.items()) == iterable_samples_before

    def test_update_sample_kwargs_and_iterably_empty_arrayset(self, subsample_writer_written_aset, iterable_subsamples, iterable_samples):
        aset = subsample_writer_written_aset
        aset.update(iterable_samples, fookwarg=iterable_subsamples)
        assert len(aset._samples) == len(iterable_samples) + 1

        assert 'fookwarg' in aset._samples
        for sample_idx in range(len(iterable_samples)):
            assert f'sample{sample_idx}' in aset._samples

    def test_update_sample_subsamples_duplicate_data_does_not_save_new(self, subsample_writer_written_aset, iterable_samples):
        aset = subsample_writer_written_aset
        aset.update(iterable_samples)
        old_specs = {}
        for sample_idx, sample_data in enumerate(iterable_samples):
            old_specs[f'sample{sample_idx}'] = aset._samples[f'sample{sample_idx}']._subsamples.copy()

        aset.update(iterable_samples)
        new_specs = {}
        for sample_idx, sample_data in enumerate(iterable_samples):
            new_specs[f'sample{sample_idx}'] = aset._samples[f'sample{sample_idx}']._subsamples.copy()
        assert old_specs == new_specs

    def test_update_sample_subsamples_context_manager(self, subsample_writer_written_aset, iterable_samples):
        aset = subsample_writer_written_aset
        assert aset._is_conman is False
        with aset as cm_aset:
            assert cm_aset._is_conman is True
            added = cm_aset.update(iterable_samples)
            assert added is None
        assert aset._is_conman is False

        assert len(aset._samples) == len(iterable_samples)
        for sample_idx, sample_data in enumerate(iterable_samples):
            assert f'sample{sample_idx}' in aset._samples

    def test_setitem_sample_subsamples_empty_arrayset(
            self, multi_item_generator, subsample_writer_written_aset, iterable_subsamples
    ):
        aset = subsample_writer_written_aset

        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = iterable_subsamples
        assert len(aset._samples) == len(iterable_subsamples)

        for sample_idx in range(multi_item_generator):
            assert f'sample{sample_idx}' in aset._samples
            assert len(aset._samples[f'sample{sample_idx}']._subsamples) == len(iterable_subsamples)
            for subsample_idx in range(len(iterable_subsamples)):
                assert f'subsample{subsample_idx}' in aset._samples[f'sample{sample_idx}']._subsamples

    def test_setitem_sample_subsamples_contextmanager(
            self, multi_item_generator, subsample_writer_written_aset, iterable_subsamples
    ):
        aset = subsample_writer_written_aset
        assert aset._is_conman is False
        with aset as aset_cm:
            assert aset_cm._is_conman is True
            for sample_idx in range(multi_item_generator):
                aset_cm[f'sample{sample_idx}'] = iterable_subsamples
            assert len(aset_cm._samples) == len(iterable_subsamples)
            assert aset_cm._samples[f'sample{sample_idx}']._is_conman is True
        assert aset._is_conman is False

        for sample_idx in range(multi_item_generator):
            assert f'sample{sample_idx}' in aset._samples
            assert len(aset._samples[f'sample{sample_idx}']._subsamples) == len(iterable_subsamples)
            for subsample_idx in range(len(iterable_subsamples)):
                assert f'subsample{subsample_idx}' in aset._samples[
                    f'sample{sample_idx}']._subsamples

    def test_update_subsamples_empty_arrayset(self, multi_item_generator, subsample_writer_written_aset, iterable_subsamples):
        aset = subsample_writer_written_aset
        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + 10}
            aset[f'sample{sample_idx}'].update(iterable_subsamples)
        assert len(aset._samples) == len(iterable_subsamples)

        for sample_idx in range(multi_item_generator):
            assert f'sample{sample_idx}' in aset._samples
            assert len(aset._samples[f'sample{sample_idx}']._subsamples) == len(iterable_subsamples) + 1
            assert 'foo' in aset._samples[f'sample{sample_idx}']._subsamples
            for subsample_idx in range(len(iterable_subsamples)):
                assert f'subsample{subsample_idx}' in aset._samples[f'sample{sample_idx}']._subsamples

    def test_update_subsamples_via_kwargs_empty_arrayset(self, multi_item_generator, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + 10}
            aset[f'sample{sample_idx}'].update(bar=np.arange(16, dtype=np.uint8).reshape(4, 4) + 20)
        assert len(aset._samples) == multi_item_generator

        for sample_idx in range(multi_item_generator):
            assert f'sample{sample_idx}' in aset._samples
            assert len(aset._samples[f'sample{sample_idx}']._subsamples) == 2
            assert 'foo' in aset._samples[f'sample{sample_idx}']._subsamples
            assert 'bar' in aset._samples[f'sample{sample_idx}']._subsamples

    def test_update_subsamples_kwargs_and_other_dict_doesnt_modify_input_in_calling_scopy(
        self, multi_item_generator, subsample_writer_written_aset, iterable_subsamples
    ):
        """ensure bug does not revert.

        Had a case where if dict was passed as ``other`` along with kwargs, the operation
        would complete as normally, but when control returned to the caller the original
        dict passed in as ``other`` would have been silently merged with the kwargs.
        """
        aset = subsample_writer_written_aset
        if not isinstance(iterable_subsamples, dict):
            return
        iterable_subsamples_before = list(iterable_subsamples.keys())

        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + 10}
            aset[f'sample{sample_idx}'].update(iterable_subsamples, kwargadded=np.arange(16, dtype=np.uint8).reshape(4, 4))
            # in bug case, would now observe that iterable_subsamples would have been
            # silently modified in a method analogous to calling:
            #
            #   ``iterable_subsamples.update({'kwargadded': np.array})``
            #
            assert list(iterable_subsamples.keys()) == iterable_subsamples_before
        assert list(iterable_subsamples.keys()) == iterable_subsamples_before

    def test_update_subsamples_via_kwargs_and_iterable_empty_arrayset(
        self, multi_item_generator, subsample_writer_written_aset, iterable_subsamples
    ):
        aset = subsample_writer_written_aset
        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + 10}
            aset[f'sample{sample_idx}'].update(iterable_subsamples, bar=np.arange(16, dtype=np.uint8).reshape(4, 4))

        assert len(aset._samples) == multi_item_generator

        for sample_idx in range(multi_item_generator):
            assert f'sample{sample_idx}' in aset._samples
            assert len(aset._samples[f'sample{sample_idx}']._subsamples) == len(iterable_subsamples) + 2
            assert 'foo' in aset._samples[f'sample{sample_idx}']._subsamples
            assert 'bar' in aset._samples[f'sample{sample_idx}']._subsamples

    @pytest.mark.parametrize('backend', fixed_shape_backend_params)
    def test_update_subsamples_context_manager(self, backend, multi_item_generator,
                                               iterable_subsamples, repo):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('foo', shape=(4, 4), dtype=np.uint8,
                                        backend=backend, contains_subsamples=True)

        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + 10}
            assert aset._is_conman is False
            with aset[f'sample{sample_idx}'] as sample_cm:
                assert sample_cm._is_conman is True
                assert aset._is_conman is True
                sample_cm.update(iterable_subsamples)
            assert aset._is_conman is False
        assert len(aset._samples) == len(iterable_subsamples)

        for sample_idx in range(multi_item_generator):
            assert f'sample{sample_idx}' in aset._samples
            assert len(aset._samples[f'sample{sample_idx}']._subsamples) == len(iterable_subsamples) + 1
            assert 'foo' in aset._samples[f'sample{sample_idx}']._subsamples
            for subsample_idx in range(len(iterable_subsamples)):
                assert f'subsample{subsample_idx}' in aset._samples[f'sample{sample_idx}']._subsamples
        co.close()

    def test_setitem_sample_empty_arrayset(self, multi_item_generator, iterable_subsamples, subsample_writer_written_aset):
        aset = subsample_writer_written_aset

        subsamples_dict = dict(iterable_subsamples)
        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + 10}
            for subsample_key, subsample_val in subsamples_dict.items():
                aset[f'sample{sample_idx}'][subsample_key] = subsample_val
        assert len(aset._samples) == len(iterable_subsamples)

        for sample_idx in range(multi_item_generator):
            assert f'sample{sample_idx}' in aset._samples
            assert len(aset._samples[f'sample{sample_idx}']._subsamples) == len(subsamples_dict) + 1
            assert 'foo' in aset._samples[f'sample{sample_idx}']._subsamples
            for subkey in subsamples_dict.keys():
                assert subkey in aset._samples[f'sample{sample_idx}']._subsamples

    def test_setitem_sample_setitem_subsample_empty_arrayset_fails(self, subsample_writer_written_aset):
        """This should fail because __getitem___ raises keyerror when

        ``aset[foo-sample][subsample] = np.ndarray`` runs.

        The ``aset[foo-sample]`` part fails with KeyError, and no subsample
        accessor is returned for the __setitem__ call following __getitem__
        """
        aset = subsample_writer_written_aset
        with pytest.raises(KeyError, match='sample'):
            aset['sample']
        with pytest.raises(KeyError, match='sample'):
            aset['sample']['subsample'] = np.arange(16, dtype=np.uint8).reshape(4, 4)
        assert len(aset) == 0

    def test_setitem_subsamples_contextmanager(self, multi_item_generator, iterable_subsamples, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        subsamples_dict = dict(iterable_subsamples)
        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + 10}
            assert aset._is_conman is False
            with aset[f'sample{sample_idx}'] as sample_cm:
                assert sample_cm._is_conman is True
                assert aset._is_conman is True
                for subsample_key, subsample_val in subsamples_dict.items():
                    sample_cm[subsample_key] = subsample_val
            assert aset._is_conman is False
        assert len(aset._samples) == len(iterable_subsamples)

        for sample_idx in range(multi_item_generator):
            assert f'sample{sample_idx}' in aset._samples
            assert len(aset._samples[f'sample{sample_idx}']._subsamples) == len(subsamples_dict) + 1
            assert 'foo' in aset._samples[f'sample{sample_idx}']._subsamples
            for subkey in subsamples_dict.keys():
                assert subkey in aset._samples[f'sample{sample_idx}']._subsamples

    def test_append_subsamples_empty_arrayset(self, multi_item_generator, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {
                'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + ((sample_idx * 2) + 1)
            }
            outkey = aset[f'sample{sample_idx}'].append(
                np.arange(16, dtype=np.uint8).reshape(4, 4) + sample_idx
            )
            assert 'foo' in aset._samples[f'sample{sample_idx}']._subsamples
            assert outkey in aset._samples[f'sample{sample_idx}']._subsamples
        assert len(aset._samples) == multi_item_generator

    def test_append_subsamples_contextmanager(self, multi_item_generator, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        for sample_idx in range(multi_item_generator):
            aset[f'sample{sample_idx}'] = {
                'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + ((sample_idx * 2) + 1)
            }
            assert aset._is_conman is False
            with aset[f'sample{sample_idx}'] as sample_cm:
                assert aset._is_conman is True
                assert sample_cm._is_conman is True
                outkey = sample_cm.append(np.arange(16, dtype=np.uint8).reshape(4, 4) + sample_idx)
            assert aset._is_conman is False
            assert 'foo' in aset._samples[f'sample{sample_idx}']._subsamples
            assert outkey in aset._samples[f'sample{sample_idx}']._subsamples
        assert len(aset._samples) == multi_item_generator

    @pytest.mark.parametrize('backend', fixed_shape_backend_params)
    @pytest.mark.parametrize('other', [
        [f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)],
        (f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)),
    ])
    def test_update_noniterable_subsample_iter_fails(self, backend, other, repo):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('foo', shape=(4, 4), dtype=np.uint8,
                                        backend=backend, contains_subsamples=True)
        aset[f'foo'] = {'foo': np.arange(16, dtype=np.uint8).reshape(4, 4) + 10}
        with pytest.raises(ValueError, match='dictionary update sequence'):
            aset['foo'].update(other)
        assert len(aset._samples) == 1
        assert len(aset._samples['foo']._subsamples) == 1
        assert 'foo' in aset._samples['foo']._subsamples
        assert 'subsample1' not in aset._samples['foo']._subsamples
        co.close()

    @pytest.mark.parametrize('backend', fixed_shape_backend_params)
    def test_update_subsamples_with_too_many_arguments_fails(self, backend, repo):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('foo', shape=(4, 4), dtype=np.uint8,
                                        backend=backend, contains_subsamples=True)
        arr = np.arange(16, dtype=np.uint8).reshape(4, 4)
        aset[f'foo'] = {'foo': arr + 10}
        with pytest.raises(TypeError, match='takes from 1 to 2 positional arguments'):
            aset['foo'].update('fail', arr)
        assert len(aset._samples) == 1
        assert len(aset._samples['foo']._subsamples) == 1
        assert 'foo' in aset._samples['foo']._subsamples
        co.close()

    @pytest.mark.parametrize('backend', fixed_shape_backend_params)
    def test_update_subsamples_with_too_few_arguments_fails(self, backend, repo):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('foo', shape=(4, 4), dtype=np.uint8,
                                        backend=backend, contains_subsamples=True)
        arr = np.arange(16, dtype=np.uint8).reshape(4, 4)
        aset[f'foo'] = {'foo': arr + 10}
        with pytest.raises(ValueError, match='dictionary update sequence element #0 has length 1; 2 is required'):
            aset['foo'].update('fail')
        assert len(aset._samples) == 1
        assert len(aset._samples['foo']._subsamples) == 1
        assert 'foo' in aset._samples['foo']._subsamples
        co.close()

    @pytest.mark.parametrize('other', [
        ['sample1', [[f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)]]],
        ['sample1', ((f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)),)],
        ('sample1', ((f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)),),),
        ('sample1', [[f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)]],),
        ['sample1', ([f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)])],
        ['sample1', [(f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4))]],
        ('sample1', ([f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)]),),
        ('sample1', [(f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4))],),
        ['sample1', [f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)]],
        ['sample1', (f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4))],
        ('sample1', [f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)],),
        ('sample1', (f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)),),
        ('sample1', {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)},),
        ['sample1', {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}],
    ])
    def test_update_noniterable_samples_fails(self, other, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        with pytest.raises(ValueError, match='dictionary update sequence'):
            aset.update(other)
        assert len(aset._samples) == 0

    @pytest.mark.parametrize('other', [
        [['sample1', [f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)]]],
        [['sample1', (f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4),)]],
        (('sample1', (f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4))),),
        (('sample1', [f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)]),),
        {'sample1': [f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)]},
        {'sample1': (f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4))},
        {'sample1': (f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4))},
        {'sample1': [f'subsample1', np.arange(16, dtype=np.uint8).reshape(4, 4)]},
    ])
    def test_update_noniterable_subsamples_fails(self, other, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        with pytest.raises(ValueError, match='dictionary update sequence'):
            aset.update(other)
        assert len(aset._samples) == 0

    @pytest.mark.parametrize('other', [
        {'sample1!': {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {-2: {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'lol cat': {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'sample 1': {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {('sample', 'one'): {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {(1, 2): {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {('sample', 2): {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {(1, 'sample'): {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
    ])
    def test_update_invalid_sample_key_fails(self, other, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        with pytest.raises(ValueError, match='is not suitable'):
            aset.update(other)
        assert len(aset._samples) == 0

    @pytest.mark.parametrize('other', [
        {'sample': {f'subsample1!': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'sample': {f'subsample 1': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'sample': {-2: np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'sample': {f'subsample1\n': np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'sample': {(1, 2): np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'sample': {('s1', 's2'): np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'sample': {('s1', 1): np.arange(16, dtype=np.uint8).reshape(4, 4)}},
        {'sample': {(1, 's1'): np.arange(16, dtype=np.uint8).reshape(4, 4)}},
    ])
    def test_update_sample_invalid_subsample_key_fails(self, other, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        with pytest.raises(ValueError, match='is not suitable'):
            aset.update(other)
        assert len(aset._samples) == 0

    @pytest.mark.parametrize('variable_shape,backend', [
        *[[True, be] for be in variable_shape_backend_params],
        *[[False, be] for be in fixed_shape_backend_params],
    ])
    @pytest.mark.parametrize('other', [
        {'sample': {f'subsample1': np.arange(9, dtype=np.uint8).reshape(3, 3)}},
        {'sample': {f'subsample1': np.arange(8, dtype=np.uint8).reshape(2, 2, 2)}},
        {'sample': {f'subsample1': np.arange(4, dtype=np.float32).reshape(2, 2)}},
        {'sample': {f'subsample1': np.arange(4, dtype=np.uint8).reshape((2, 2), order='F')}},
        {'sample': {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4).tolist()}},
    ])
    def test_update_sample_invalid_array_fails_fixed_shape(self, backend, variable_shape, other, repo):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('foo',
                                        shape=(2, 2), dtype=np.uint8, variable_shape=variable_shape,
                                        backend=backend, contains_subsamples=True)
        with pytest.raises(ValueError):
            aset.update(other)
        assert len(aset._samples) == 0
        co.close()

    @pytest.mark.parametrize('other', [
        {f'subsample1!': np.arange(16, dtype=np.uint8).reshape(4, 4)},
        {f'subsample 1': np.arange(16, dtype=np.uint8).reshape(4, 4)},
        {-2: np.arange(16, dtype=np.uint8).reshape(4, 4)},
        {f'subsample1\n': np.arange(16, dtype=np.uint8).reshape(4, 4)},
        {(1, 2): np.arange(16, dtype=np.uint8).reshape(4, 4)},
        {('s1', 's2'): np.arange(16, dtype=np.uint8).reshape(4, 4)},
        {('s1', 1): np.arange(16, dtype=np.uint8).reshape(4, 4)},
        {(1, 's1'): np.arange(16, dtype=np.uint8).reshape(4, 4)},
    ])
    def test_update_subsample_invalid_subsample_key_fails(self, other, subsample_writer_written_aset):
        aset = subsample_writer_written_aset
        aset['sample'] = {0: np.zeros((4, 4), dtype=np.uint8)}
        with pytest.raises(ValueError, match='is not suitable'):
            aset['sample'].update(other)
        assert len(aset._samples) == 1
        assert len(aset._samples['sample']._subsamples) == 1
        assert 0 in aset._samples['sample']._subsamples

    @pytest.mark.parametrize('variable_shape,backend', [
        *[[False, be] for be in fixed_shape_backend_params],
    ])
    @pytest.mark.parametrize('other', [
        {f'subsample1': np.arange(9, dtype=np.uint8).reshape(3, 3)},
        {f'subsample1': np.arange(8, dtype=np.uint8).reshape(2, 2, 2)},
        {f'subsample1': np.arange(4, dtype=np.float32).reshape(2, 2)},
        {f'subsample1': np.arange(4, dtype=np.uint8).reshape((2, 2), order='F')},
        {f'subsample1': np.arange(16, dtype=np.uint8).reshape(4, 4).tolist()},
    ])
    def test_update_subsample_invalid_array_fails_fixed_shape(self, backend, variable_shape, other, repo):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('foo',
                                        shape=(4, 4), dtype=np.uint8, variable_shape=variable_shape,
                                        backend=backend, contains_subsamples=True)
        aset['sample'] = {0: np.zeros((4, 4), dtype=np.uint8)}
        with pytest.raises(ValueError):
            aset['sample'].update(other)
        assert len(aset._samples) == 1
        assert len(aset._samples['sample']._subsamples) == 1
        assert 0 in aset._samples['sample']._subsamples
        co.close()


# --------------------------- Test Remove Data -------------------------------------


@pytest.fixture(scope='class')
def subsample_data_map():
    arr = np.arange(5*7).astype(np.uint16).reshape((5, 7))
    res = {
        'foo': {
            0: arr,
            1: arr + 1,
            2: arr + 2
        },
        2: {
            'bar': arr + 3,
            'baz': arr + 4
        }
    }
    return res


@pytest.fixture(params=fixed_shape_backend_params, scope='class')
def backend_param(request):
    return request.param


@pytest.fixture(params=[False, True], scope='class')
def write_enabled(request):
    return request.param


@pytest.fixture(scope='class')
def initialized_arrayset(write_enabled, backend_param, classrepo, subsample_data_map):
    co = classrepo.checkout(write=True)
    aset = co.add_ndarray_column(f'foo{backend_param}{int(write_enabled)}',
                                    shape=(5, 7), dtype=np.uint16, backend=backend_param,
                                    contains_subsamples=True)
    aset.update(subsample_data_map)
    co.commit(f'done {backend_param}{write_enabled}')
    co.close()
    if write_enabled:
        nco = classrepo.checkout(write=True)
        yield nco.columns[f'foo{backend_param}{int(write_enabled)}']
        nco.close()
    else:
        nco = classrepo.checkout()
        yield nco.columns[f'foo{backend_param}{int(write_enabled)}']
        nco.close()


@pytest.fixture()
def initialized_arrayset_write_only(backend_param, repo, subsample_data_map):
    co = repo.checkout(write=True)
    aset = co.add_ndarray_column('foo', shape=(5, 7), dtype=np.uint16,
                                    backend=backend_param, contains_subsamples=True)
    aset.update(subsample_data_map)
    yield co.columns['foo']
    co.close()


class TestRemoveData:

    # --------------------- delete -----------------------------

    def test_delitem_single_sample_from_arrayset(self, initialized_arrayset_write_only):
        aset = initialized_arrayset_write_only
        del aset['foo']
        assert 'foo' not in aset._samples
        assert 'foo' not in aset

    def test_delitem_single_subsample_from_sample(self, initialized_arrayset_write_only):
        aset = initialized_arrayset_write_only
        del aset['foo'][0]
        assert 0 not in aset._samples['foo']._subsamples
        assert 0 not in aset['foo']

    def test_delitem_sample_nonexisting_keys_fails(self, initialized_arrayset_write_only):
        aset = initialized_arrayset_write_only
        assert 'doesnotexist' not in aset._samples
        assert 'doesnotexist' not in aset
        with pytest.raises(KeyError):
            del aset['doesnotexist']

    def test_delitem_single_subsample_nonexisting_key_fails(self, initialized_arrayset_write_only):
        aset = initialized_arrayset_write_only
        assert 'foo' in aset._samples
        assert 'foo' in aset
        assert 'doesnotexist' not in aset._samples['foo']._subsamples
        assert 'doesnotexist' not in aset['foo']
        with pytest.raises(KeyError):
            del aset['foo']['doesnotexist']

    def test_delitem_multiple_samples_fails_keyerror(self, initialized_arrayset_write_only):
        aset = initialized_arrayset_write_only
        with pytest.raises(KeyError, match="('foo', 2)"):
            del aset['foo', 2]
        assert 'foo' in aset
        assert 2 in aset

    # ------------------------ pop ----------------------------

    def test_pop_single_sample_from_arrayset(self, initialized_arrayset_write_only, subsample_data_map):
        aset = initialized_arrayset_write_only
        res = aset.pop('foo')
        assert 'foo' not in aset
        assert isinstance(res, dict)
        assert len(res) == len(subsample_data_map['foo'])
        for expected_k, expected_v in subsample_data_map['foo'].items():
            assert_equal(res[expected_k], expected_v)

    def test_pop_multiple_samples_from_arrayset_fails(self, initialized_arrayset_write_only):
        aset = initialized_arrayset_write_only
        with pytest.raises(TypeError, match="takes 2 positional arguments but 3 were"):
            aset.pop('foo', 2)
        assert 'foo' in aset
        assert 2 in aset

    def test_pop_single_subsample_from_sample(self, initialized_arrayset_write_only, subsample_data_map):
        aset = initialized_arrayset_write_only
        res = aset['foo'].pop(0)
        assert 0 not in aset['foo']
        assert isinstance(res, np.ndarray)
        assert_equal(res, subsample_data_map['foo'][0])

    def test_pop_multiple_subsample_from_sample_fails(self, initialized_arrayset_write_only):
        aset = initialized_arrayset_write_only
        with pytest.raises(TypeError, match="takes 2 positional arguments but 3 were given"):
            aset['foo'].pop(*[0, 1])
        assert 0 in aset['foo']
        assert 1 in aset['foo']


# ------------------------------ Container Introspection -----------------------------------


class TestContainerIntrospection:

    def test_get_sample_returns_object(self, initialized_arrayset, subsample_data_map):
        from hangar.columns.layout_nested import FlatSubsampleReader, NestedSampleReader

        aset = initialized_arrayset
        assert isinstance(aset, NestedSampleReader)
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert isinstance(sample, FlatSubsampleReader)

    # -------------------------- test __dunder__ methods ----------------------------------

    def test_get_sample_test_subsample_len_method(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert len(sample) == len(subsample_data)

    def test_get_sample_test_subsample_contains_method(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            for subsample_name in subsample_data.keys():
                assert subsample_name in sample

    def test_sample_len_reported_correctly(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        assert len(aset) == len(subsample_data_map)
        assert aset.num_subsamples == sum([len(subsample) for subsample in subsample_data_map.values()])

    # ----------------------------- test property ---------------------------

    def test_get_sample_test_subsample_sample_property(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert sample.sample == sample_name

    def test_get_sample_test_subsample_arrayset_property(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert sample.column.startswith('foo')

    def test_get_sample_test_data_property(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            res = sample.data
            assert isinstance(res, dict)
            assert len(res) == len(subsample_data)
            for k, v in res.items():
                assert_equal(v, subsample_data[k])

    def test_get_sample_test_subsample_contains_remote_references_property(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        # test works before add remote references
        aset.contains_remote_references is False
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert sample.contains_remote_references is False

        # add subsamples which are not local to each subsample
        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        aset['foo']._subsamples[50] = template
        aset[2]._subsamples[50] = template

        aset.contains_remote_references is True
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert sample.contains_remote_references is True

        del aset._samples['foo']._subsamples[50]
        del aset._samples[2]._subsamples[50]

    def test_get_sample_test_subsample_remote_reference_keys_property(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        # test works before add remote references
        assert aset.remote_reference_keys == ()
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert sample.remote_reference_keys == ()

        # add subsamples which are not local to each subsample
        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        aset['foo']._subsamples[50] = template
        aset[2]._subsamples[50] = template

        assert aset.remote_reference_keys == (2, 'foo') or ('foo', 2)
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert sample.remote_reference_keys == (50,)

        del aset._samples['foo']._subsamples[50]
        del aset._samples[2]._subsamples[50]

    def test_getattr_does_not_raise_permission_error_if_alive(self, initialized_arrayset):
        aset = initialized_arrayset

        assert hasattr(aset, 'doesnotexist') is False  # does not raise error
        assert hasattr(aset, '_mode') is True
        with pytest.raises(AttributeError):
            assert getattr(aset, 'doesnotexist')
        assert getattr(aset, '_mode') == 'a' if aset.iswriteable else 'r'

        sample = aset['foo']
        assert hasattr(sample, 'doesnotexist') is False  # does not raise error
        assert hasattr(sample, '_mode') is True
        with pytest.raises(AttributeError):
            assert getattr(sample, 'doesnotexist')
        assert getattr(sample, '_mode') == 'a' if aset.iswriteable else 'r'

        # mock up destruct call in sample and aset.
        original = getattr(aset, '_mode')
        delattr(aset, '_mode')
        delattr(sample, '_mode')
        with pytest.raises(PermissionError):
            hasattr(aset, 'doesnotexist')
        with pytest.raises(PermissionError):
            hasattr(aset, '_mode')

        with pytest.raises(PermissionError):
            hasattr(sample, 'doesnotexist')
        with pytest.raises(PermissionError):
            hasattr(sample, '_mode')
        setattr(aset, '_mode', original)
        setattr(sample, '_mode', original)

# ------------------------------ Getting Data --------------------------------------------


class TestGetDataMethods:

    def test_get_sample_missing_key(self, initialized_arrayset):
        aset = initialized_arrayset
        returned = aset.get('doesnotexist')
        assert returned is None
        default_returned = aset.get(9999, default=True)
        assert default_returned is True

    def test_getitem_sample_missing_key(self, initialized_arrayset):
        aset = initialized_arrayset
        with pytest.raises(KeyError):
            aset['doesnotexist']

    def test_get_sample_get_subsample(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            for subsample_name, subsample_value in subsample_data.items():
                res = sample.get(subsample_name)
                assert_equal(res, subsample_value)

    def test_getitem_sample_getitem_subsample(self, initialized_arrayset, subsample_data_map):
        from hangar.columns.layout_nested import FlatSubsampleReader

        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset[sample_name]
            assert isinstance(sample, FlatSubsampleReader)
            for subsample_name, subsample_value in subsample_data.items():
                res = sample[subsample_name]
                assert_equal(res, subsample_value)

    def test_get_sample_get_subsample_missing_key(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name in subsample_data_map.keys():
            sample = aset.get(sample_name)
            returned = sample.get('doesnotexist')
            assert returned is None
            default_returned = sample.get(9999, default=True)
            assert default_returned is True

    def test_getitem_sample_getitem_subsample_missing_key(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name in subsample_data_map.keys():
            sample = aset[sample_name]
            with pytest.raises(KeyError):
                sample['doesnotexist']

    def test_get_sample_get_multiple_subsamples_fails(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            with pytest.raises(TypeError):
                sample.get(*list(list(subsample_data.keys())[:2]), default=None)

    def test_get_sample_getitem_single_subsample(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            for subsample_name, subsample_value in subsample_data.items():
                res = sample[subsample_name]
                assert_equal(res, subsample_value)

    def test_get_sample_getitem_single_subsample_missing_key(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name in subsample_data_map.keys():
            sample = aset.get(sample_name)
            returned = sample.get('doesnotexist')
            assert returned is None
            default_returned = sample.get(9999, default=True)
            assert default_returned is True

    def test_get_sample_getitem_multiple_subsamples_fails(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            with pytest.raises(TypeError):
                sample[list(subsample_data.keys())[:2]]

    def test_get_sample_getitem_subsamples_with_ellipsis(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            res = sample[...]
            assert isinstance(res, dict)
            assert len(res) == len(subsample_data)
            for k, v in res.items():
                assert_equal(v, subsample_data[k])

    def test_get_sample_getitem_subsamples_with_keys_and_ellipsis_fails(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            existing_subsample_key = next(iter(subsample_data.keys()))
            with pytest.raises(TypeError):
                sample[..., existing_subsample_key]
            with pytest.raises(TypeError):
                sample[..., [existing_subsample_key]]

    def test_get_sample_getitem_subsamples_with_unbound_slice(self, initialized_arrayset, subsample_data_map):
        """unbound slice is ``slice(None) == [:]``"""
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            res = sample[:]
            assert isinstance(res, dict)
            assert len(res) == len(subsample_data)
            for k, v in res.items():
                assert_equal(v, subsample_data[k])

    def test_get_sample_getitem_subsamples_with_bounded_slice(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            res = sample[0:2]
            assert isinstance(res, dict)
            assert len(res) == 2
            for k, v in res.items():
                assert_equal(v, subsample_data[k])

    def test_get_sample_getitem_subsamples_with_out_of_bounds_slice_does_not_fail(
            self, initialized_arrayset, subsample_data_map):
        """Odd python behavior we emulate: out of bounds sequence slicing is allowed.

        Instead of throwing an exception, the slice is treated as if it should just
        go up to the total number of elements in the container. For example:
            [1, 2, 3][0:5] == [1, 2, 3]
        """
        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            res = sample[0:5]
            assert isinstance(res, dict)
            assert len(res) == len(subsample_data)
            for k, v in res.items():
                assert_equal(v, subsample_data[k])

    def test_aset_contextmanager(self, initialized_arrayset, subsample_data_map):
        assert initialized_arrayset._is_conman is False
        with initialized_arrayset as aset:
            assert aset._is_conman is True
            for sample_name, subsample_data in subsample_data_map.items():
                sample = aset.get(sample_name)
                assert sample._is_conman is True
                for subsample_name, expected_val in subsample_data.items():
                    assert_equal(sample.get(subsample_name), expected_val)
                assert sample._is_conman is True
        assert initialized_arrayset._is_conman is False
        assert aset._is_conman is False
        assert sample._is_conman is False

    def test_sample_contextmanager(self, initialized_arrayset, subsample_data_map):
        for sample_name, subsample_data in subsample_data_map.items():
            sample = initialized_arrayset.get(sample_name)
            assert initialized_arrayset._is_conman is False
            assert sample._is_conman is False
            with sample as sample_cm:
                assert sample_cm._is_conman is True
                assert initialized_arrayset._is_conman is True
                for subsample_name, expected_val in subsample_data.items():
                    assert_equal(sample_cm.get(subsample_name), expected_val)
            assert sample._is_conman is False
            assert initialized_arrayset._is_conman is False
        assert initialized_arrayset._is_conman is False
        assert sample._is_conman is False

    def test_sample_subsample_contextmanager(self, initialized_arrayset, subsample_data_map):
        assert initialized_arrayset._is_conman is False
        with initialized_arrayset as aset:
            assert aset._is_conman is True
            assert aset._enter_count == 1
            for sample_name, subsample_data in subsample_data_map.items():
                sample = aset.get(sample_name)
                assert sample._is_conman is True
                assert sample._enter_count == 1
                with sample as sample_cm:
                    assert aset._is_conman is True
                    assert sample_cm._is_conman is True
                    assert aset._enter_count == 2
                    assert sample_cm._enter_count == 2
                    for subsample_name, expected_val in subsample_data.items():
                        assert_equal(sample_cm.get(subsample_name), expected_val)
                assert aset._is_conman is True
                assert sample_cm._is_conman is True
                assert aset._enter_count == 1
                assert sample_cm._enter_count == 1
        assert initialized_arrayset._is_conman is False
        assert aset._is_conman is False
        assert sample._is_conman is False
        assert aset._enter_count == 0
        assert sample_cm._enter_count == 0

    def test_sample_reentrant_contextmanager_fails(self, initialized_arrayset, subsample_data_map):
        assert initialized_arrayset._is_conman is False

        with initialized_arrayset as aset:
            assert aset._is_conman is True
            assert aset._enter_count == 1
            for sample_name, subsample_data in subsample_data_map.items():
                sample = aset.get(sample_name)
                assert sample._is_conman is True
                assert sample._enter_count == 1
                with sample as sample_cm:
                    assert aset._is_conman is True
                    assert sample_cm._is_conman is True
                    assert aset._enter_count == 2
                    assert sample_cm._enter_count == 2
                    for subsample_name, expected_val in subsample_data.items():
                        assert_equal(sample_cm.get(subsample_name), expected_val)
                # reentrant demonstrated here here
                with sample as sample_cm2:
                    assert aset._is_conman is True
                    assert sample_cm._is_conman is True
                    assert sample_cm2._is_conman is True
                    assert aset._enter_count == 2
                    assert sample_cm._enter_count == 2
                    assert sample_cm2._enter_count == 2
                    for subsample_name, expected_val in subsample_data.items():
                        assert_equal(sample_cm2.get(subsample_name), expected_val)
                assert aset._is_conman is True
                assert sample_cm._is_conman is True
                assert aset._enter_count == 1
                assert sample_cm._enter_count == 1
        assert initialized_arrayset._is_conman is False
        assert aset._is_conman is False
        assert sample._is_conman is False
        assert aset._enter_count == 0
        assert sample_cm._enter_count == 0

    # -------------------------- dict-style iteration methods ---------------------------

    def test_calling_iter_on_arrayset(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        arrayset_it = iter(aset)  # returns iterator over sample keys
        for sample_name in arrayset_it:
            assert sample_name in aset
            assert sample_name in subsample_data_map

    def test_calling_iter_on_sample_in_arrayset(self, initialized_arrayset, subsample_data_map):
        aset = initialized_arrayset
        arrayset_it = iter(aset)  # returns iterator over sample keys
        for sample_name in arrayset_it:
            assert sample_name in aset
            assert sample_name in subsample_data_map

            sample_it = iter(aset[sample_name])  # returns iterator over subsample keys
            for subsample_name in sample_it:
                assert subsample_name in aset[sample_name]
                assert subsample_name in subsample_data_map[sample_name]

    def test_get_sample_keys_method(self, initialized_arrayset):
        from collections.abc import Iterator
        aset = initialized_arrayset

        assert isinstance(aset.keys(), Iterator)
        res = list(aset.keys())
        assert len(res) == 2
        assert 2 and 'foo' in res

    def test_get_sample_keys_method_local_only(self, initialized_arrayset):
        from collections.abc import Iterator
        aset = initialized_arrayset

        # add subsamples which are not local to each subsample
        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        aset['foo']._subsamples[50] = template

        assert isinstance(aset.keys(local=True), Iterator)
        res = list(aset.keys(local=True))
        assert len(res) == 1
        assert 2 in res

        del aset._samples['foo']._subsamples[50]

    def test_get_sample_subsample_keys_method(self, initialized_arrayset, subsample_data_map):
        from collections.abc import Iterator

        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert isinstance(sample.keys(), Iterator)
            res = list(sample.keys())
            for k in res:
                assert k in subsample_data

    def test_get_sample_subsample_keys_method_local_only(self, initialized_arrayset, subsample_data_map):
        from collections.abc import Iterator
        aset = initialized_arrayset

        # add subsamples which are not local to each subsample
        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        aset['foo']._subsamples[50] = template
        aset[2]._subsamples[50] = template

        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)

            # test local only properties
            assert isinstance(sample.keys(local=True), Iterator)
            res = list(sample.keys(local=True))
            assert len(res) == len(subsample_data)
            for k in res:
                assert k in subsample_data
                assert k != 50

            # compare to local+remote properties
            assert isinstance(sample.keys(local=False), Iterator)
            res = list(sample.keys(local=False))
            assert len(res) == len(subsample_data) + 1
            assert 50 in res
            for k in res:
                assert k in list(subsample_data.keys()) + [50]

        del aset._samples['foo']._subsamples[50]
        del aset._samples[2]._subsamples[50]

    def test_get_sample_values_method(self, initialized_arrayset):
        from hangar.columns.layout_nested import FlatSubsampleReader
        from collections.abc import Iterator
        aset = initialized_arrayset

        assert isinstance(aset.values(), Iterator)
        res = list(aset.values())
        assert len(res) == 2
        for sample in res:
            assert sample.sample == 'foo' or 2
            assert isinstance(sample, FlatSubsampleReader)

    def test_get_sample_values_method_local_only(self, initialized_arrayset):
        from hangar.columns.layout_nested import FlatSubsampleReader
        from collections.abc import Iterator
        aset = initialized_arrayset
        # add subsamples which are not local to each subsample
        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        aset['foo']._subsamples[50] = template

        assert isinstance(aset.values(local=True), Iterator)
        res = list(aset.values(local=True))
        assert len(res) == 1
        sample = res[0]
        assert sample.sample == 2
        assert isinstance(sample, FlatSubsampleReader)

        del aset._samples['foo']._subsamples[50]

    def test_get_sample_subsample_values_method(self, initialized_arrayset, subsample_data_map):
        from collections.abc import Iterator

        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert isinstance(sample.values(), Iterator)
            res = list(sample.values())
            for v in res:
                assert any([np.allclose(v, arr) for arr in subsample_data.values()])

    def test_get_sample_subsample_values_method_local_only(self, initialized_arrayset, subsample_data_map):
        from collections.abc import Iterator
        aset = initialized_arrayset

        # add subsamples which are not local to each subsample
        # perform the mock
        from hangar.backends import backend_decoder
        from hangar.columns.common import open_file_handles
        template = backend_decoder(b'50:daeaaeeaebv')
        aset['foo']._subsamples[50] = template
        aset[2]._subsamples[50] = template
        mocked_fhand = open_file_handles(
            ['50'], path=initialized_arrayset._path, mode='a', schema=aset._schema)
        aset._be_fs['50'] = mocked_fhand['50']

        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)

            # test local only properties
            assert isinstance(sample.values(local=True), Iterator)
            res = list(sample.values(local=True))
            assert len(res) == len(subsample_data)
            for v in res:
                assert any([np.allclose(v, arr) for arr in subsample_data.values()])

            # test local+remote properties
            with pytest.raises(FileNotFoundError):
                list(sample.values(local=False))

        del aset._be_fs['50']
        del aset._samples['foo']._subsamples[50]
        del aset._samples[2]._subsamples[50]

    def test_get_sample_items_method(self, initialized_arrayset):
        from hangar.columns.layout_nested import FlatSubsampleReader
        from collections.abc import Iterator
        aset = initialized_arrayset

        assert isinstance(aset.items(), Iterator)
        res = list(aset.items())
        assert len(res) == 2
        for sample_name, sample in res:
            assert sample_name == 2 or 'foo'
            assert isinstance(sample, FlatSubsampleReader)
            assert sample_name == sample.sample

    def test_get_sample_items_method_local_only(self, initialized_arrayset):
        from hangar.columns.layout_nested import FlatSubsampleReader
        from collections.abc import Iterator
        aset = initialized_arrayset
        # add subsamples which are not local to each subsample
        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        aset['foo']._subsamples[50] = template

        assert isinstance(aset.items(local=True), Iterator)
        res = list(aset.items(local=True))
        assert len(res) == 1
        sample_name, sample = res[0]
        assert sample_name == 2
        assert isinstance(sample, FlatSubsampleReader)
        assert sample.sample == sample_name

        del aset._samples['foo']._subsamples[50]

    def test_get_sample_subsample_items_method(self, initialized_arrayset, subsample_data_map):
        from collections.abc import Iterator

        aset = initialized_arrayset
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            assert isinstance(sample.items(), Iterator)
            res = list(sample.items())
            for k, v in res:
                assert_equal(v, subsample_data[k])

    def test_get_sample_subsample_items_method_local_only(self, initialized_arrayset, subsample_data_map):
        from collections.abc import Iterator
        aset = initialized_arrayset

        # add subsamples which are not local to each subsample ato perform the mock
        from hangar.backends import backend_decoder
        from hangar.columns.common import open_file_handles
        template = backend_decoder(b'50:daeaaeeaebv')
        aset['foo']._subsamples[50] = template
        aset[2]._subsamples[50] = template
        mocked_fhand = open_file_handles(
            ['50'], path=initialized_arrayset._path, mode='a', schema=aset._schema)
        aset._be_fs['50'] = mocked_fhand['50']

        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)

            # test local only properties
            assert isinstance(sample.items(local=True), Iterator)
            res = list(sample.items(local=True))
            assert len(res) == len(subsample_data)
            for k, v in res:
                assert_equal(v, subsample_data[k])
                assert k != 50

            # test local+remote properties
            with pytest.raises(FileNotFoundError):
                list(sample.items(local=False))

        del aset._be_fs['50']
        del aset._samples['foo']._subsamples[50]
        del aset._samples[2]._subsamples[50]

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset3_backend", fixed_shape_backend_params)
    def test_arrayset_remote_references_property_with_none(
            self, aset1_backend, aset2_backend, aset3_backend, repo, randomsizedarray
    ):
        co = repo.checkout(write=True)
        aset1 = co.add_ndarray_column('aset1', prototype=randomsizedarray,
                                         backend=aset1_backend, contains_subsamples=True)
        aset2 = co.add_ndarray_column('aset2', shape=(2, 2), dtype=np.int,
                                         backend=aset2_backend, contains_subsamples=True)
        aset3 = co.add_ndarray_column('aset3', shape=(3, 4), dtype=np.float32,
                                         backend=aset3_backend, contains_subsamples=True)
        with aset1 as d1, aset2 as d2, aset3 as d3:
            d1[1] = {11: randomsizedarray}
            d2[1] = {21: np.ones((2, 2), dtype=np.int)}
            d3[1] = {31: np.ones((3, 4), dtype=np.float32)}

        assert co.columns.contains_remote_references == {'aset1': False, 'aset2': False, 'aset3': False}
        assert co.columns.remote_sample_keys == {'aset1': (), 'aset2': (), 'aset3': ()}
        co.close()

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset3_backend", fixed_shape_backend_params)
    def test_arrayset_remote_references_property_with_remotes(
            self, aset1_backend, aset2_backend, aset3_backend, repo, randomsizedarray
    ):
        co = repo.checkout(write=True)
        aset1 = co.add_ndarray_column('aset1', prototype=randomsizedarray,
                                         backend=aset1_backend, contains_subsamples=True)
        aset2 = co.add_ndarray_column('aset2', shape=(2, 2), dtype=np.int,
                                         backend=aset2_backend, contains_subsamples=True)
        aset3 = co.add_ndarray_column('aset3', shape=(3, 4), dtype=np.float32,
                                         backend=aset3_backend, contains_subsamples=True)
        with aset1 as d1, aset2 as d2, aset3 as d3:
            d1[1] = {11: randomsizedarray}
            d2[1] = {21: np.ones((2, 2), dtype=np.int)}
            d3[1] = {31: np.ones((3, 4), dtype=np.float32)}

        assert co.columns.contains_remote_references == {'aset1': False, 'aset2': False, 'aset3': False}
        assert co.columns.remote_sample_keys == {'aset1': (), 'aset2': (), 'aset3': ()}
        co.commit('hello')
        co.close()
        co = repo.checkout()
        # perform the mock
        # perform the mock
        from hangar.backends import backend_decoder
        template = backend_decoder(b'50:daeaaeeaebv')
        co._columns._columns['aset1']._samples[1]._subsamples[12] = template
        co._columns._columns['aset2']._samples[1]._subsamples[22] = template

        assert co.columns.contains_remote_references == {'aset1': True, 'aset2': True, 'aset3': False}
        assert co.columns.remote_sample_keys == {'aset1': (1,), 'aset2': (1,), 'aset3': ()}
        co.close()


class TestWriteThenReadCheckout:

    @pytest.mark.parametrize('backend', fixed_shape_backend_params)
    def test_add_data_commit_checkout_read_only_contains_same(self, backend, repo, subsample_data_map):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('foo', shape=(5, 7), dtype=np.uint16,
                                        backend=backend, contains_subsamples=True)
        added = aset.update(subsample_data_map)
        for sample_name, subsample_data in subsample_data_map.items():
            sample = aset.get(sample_name)
            for subsample_name, subsample_val in subsample_data.items():
                assert_equal(sample[subsample_name], subsample_val)
        co.commit('first')
        co.close()

        rco = repo.checkout()
        naset = rco.columns['foo']
        for sample_name, subsample_data in subsample_data_map.items():
            sample = naset.get(sample_name)
            for subsample_name, subsample_val in subsample_data.items():
                assert_equal(sample[subsample_name], subsample_val)
        rco.close()



import math
import struct
from typing import NamedTuple, List

import blosc
import numpy as np

from . import hangar_service_pb2
from ..utils import set_blosc_nthreads

set_blosc_nthreads()


def chunk_bytes(bytesData):
    """Slice a bytestring into subelements and store the data in a list

    Arguments
    ---------
        bytesData : bytes
            bytestring buffer of the array data

    Yields
    ------
    bytes
        data split into 32kb chunk sizes.
    """
    chunkSize = 32_000
    numIters = math.ceil(len(bytesData) / chunkSize)
    currentStart = 0
    currentEnd = chunkSize
    for i in range(numIters):
        yield bytesData[currentStart:currentEnd]
        currentStart += chunkSize
        currentEnd += chunkSize


def clientCommitChunkedIterator(commit: str, parentVal: bytes, specVal: bytes,
                                refVal: bytes) -> hangar_service_pb2.PushCommitRequest:
    """Generator splitting commit specs into chunks sent from client to server

    Parameters
    ----------
    commit : str
        commit hash which is being sent
    parentVal : bytes
        bytes representing the commits immediate parents
    specVal : bytes
        bytes representing the commit message/user specifications
    refVal : bytes
        bytes containing all records stored in the repository

    Yields
    ------
    hangar_service_pb2.PushCommitRequest
        Chunked generator of the PushCommitRequest protobuf.
    """
    commit_proto = hangar_service_pb2.CommitRecord(
        parent=parentVal,
        spec=specVal)
    byteSize = len(refVal)
    chunkIterator = chunk_bytes(refVal)
    for refChunk in chunkIterator:
        commit_proto.ref = refChunk
        request = hangar_service_pb2.PushCommitRequest(
            commit=commit,
            total_byte_size=byteSize,
            record=commit_proto)
        yield request


def tensorChunkedIterator(buf, uncomp_nbytes, itemsize, pb2_request, *, err=None):

    compBytes = blosc.compress(
        buf, clevel=3, cname='lz4', typesize=1, shuffle=blosc.NOSHUFFLE)

    request = pb2_request(
        comp_nbytes=len(compBytes),
        uncomp_nbytes=uncomp_nbytes,
        error=err)
    chunkIterator = chunk_bytes(compBytes)
    for dchunk in chunkIterator:
        request.raw_data = dchunk
        yield request


def missingHashIterator(commit, hash_bytes, err, pb2_func):
    comp_bytes = blosc.compress(
        hash_bytes, cname='zlib', clevel=3, typesize=1, shuffle=blosc.SHUFFLE)

    rpc_method = pb2_func(
        commit=commit,
        total_byte_size=len(comp_bytes),
        error=err)

    chunkIterator = chunk_bytes(comp_bytes)
    for bchunk in chunkIterator:
        rpc_method.hashs = bchunk
        yield rpc_method


def missingHashRequestIterator(commit, hash_bytes, pb2_func):
    comp_bytes = blosc.compress(
        hash_bytes, cname='zlib', clevel=3, typesize=1, shuffle=blosc.SHUFFLE)

    rpc_method = pb2_func(
        commit=commit,
        total_byte_size=len(comp_bytes))

    chunkIterator = chunk_bytes(comp_bytes)
    for bchunk in chunkIterator:
        rpc_method.hashs = bchunk
        yield rpc_method


# ------------------------ serialization formats -------------------------


ArrayIdent = NamedTuple('ArrayIdent', [
    ('digest', str),
    ('schema', str)
])

ArrayRecord = NamedTuple('ArrayRecord', [
    ('array', np.ndarray),
    ('digest', str),
    ('schema', str),
])


def serialize_arr(arr: np.ndarray) -> bytes:
    """
    dtype_num ndim dim1_size dim2_size ... dimN_size array_bytes
    """
    domain = struct.pack(f'<II', arr.dtype.num, arr.ndim)
    dims = [struct.pack(f'<I', dim) for dim in arr.shape]
    return b''.join([domain, *dims, arr.tobytes()])


def deserialize_arr(raw: bytes) -> np.ndarray:
    dtnum, ndim = struct.unpack('<II', raw[:8])
    dtype = np.typeDict[dtnum]
    dataStart = 8 + (ndim * 4)
    shape = struct.unpack(f'<{ndim}I', raw[8:dataStart])
    arr = np.frombuffer(raw, dtype=dtype, offset=dataStart).reshape(shape)
    return arr


def serialize_ident(digest: str, schema: str) -> bytes:
    """
    len_digest digest_str len_schema schema_str
    """
    ident_digest = struct.pack(f'<I{len(digest)}s', len(digest), digest.encode())
    ident_schema = struct.pack(f'<I{len(schema)}s', len(schema), schema.encode())
    return b''.join([ident_digest, ident_schema])


def deserialize_ident(raw: bytes) -> ArrayIdent:
    digestLen = struct.unpack('<I', raw[:4])[0]
    digestEnd = 4 + (digestLen * 1)
    schemaStart = 4 + digestEnd
    schemaLen = struct.unpack('<I', raw[digestEnd:schemaStart])[0]
    digest = struct.unpack(f'<{digestLen}s', raw[4:digestEnd])[0].decode()
    schema = struct.unpack(f'<{schemaLen}s', raw[schemaStart:])[0].decode()
    return ArrayIdent(digest, schema)


def serialize_record(arr: np.ndarray, digest: str, schema: str) -> bytes:
    """
    len_raw_ident len_raw_arr raw_ident, raw_arr
    """
    raw_arr = serialize_arr(arr)
    raw_ident = serialize_ident(digest, schema)
    record = struct.pack(f'<2Q', len(raw_ident), len(raw_arr))
    return b''.join([record, raw_ident, raw_arr])


def deserialize_record(raw: bytes) -> ArrayRecord:
    identStart = 16  # 2 * 8 bytes
    identLen, arrLen = struct.unpack(f'<2Q', raw[:identStart])
    identEnd = identStart + identLen
    arrEnd = identEnd + arrLen

    arr = deserialize_arr(raw[identEnd:arrEnd])
    ident = deserialize_ident(raw[identStart:identEnd])
    return ArrayRecord(arr, ident.digest, ident.schema)


def serialize_record_pack(records: List[bytes]) -> bytes:
    """
    num_records len_rec1 raw_rec1 len_rec2 raw_rec2 ... len_recN raw_recN
    """
    raw_num_records = struct.pack(f'<I', len(records))
    raw_records = [b''.join([struct.pack(f'<Q', len(rec)), rec]) for rec in records]
    return b''.join([raw_num_records, *raw_records])


def deserialize_record_pack(raw: bytes) -> List[bytes]:
    numRecords = struct.unpack(f'<I', raw[:4])[0]
    cursorPos, recs = 4, []
    for i in range(numRecords):
        lenRec = struct.unpack(f'<Q', raw[cursorPos:cursorPos+8])[0]
        recs.append(raw[cursorPos+8:cursorPos+8+lenRec])
        cursorPos += (8 + lenRec)
    return recs



import time
import shutil
import random
from random import randint
import platform
from os.path import join as pjoin
from os import mkdir

import pytest
import numpy as np

from hangar import Repository
from hangar.checkout import WriterCheckout

import hangar


variable_shape_backend_params = ['00', '10']
fixed_shape_backend_params = ['00', '01', '10']


@pytest.fixture(scope='class')
def classrepo(tmp_path_factory) -> Repository:
    old00_count = hangar.backends.hdf5_00.COLLECTION_COUNT
    old00_size = hangar.backends.hdf5_00.COLLECTION_SIZE
    old01_count = hangar.backends.hdf5_01.COLLECTION_COUNT
    old01_size = hangar.backends.hdf5_01.COLLECTION_SIZE
    old10_size = hangar.backends.numpy_10.COLLECTION_SIZE
    old30_lmdb_settings = hangar.backends.lmdb_30.LMDB_SETTINGS
    hangar.backends.hdf5_00.COLLECTION_COUNT = 5
    hangar.backends.hdf5_00.COLLECTION_SIZE = 20
    hangar.backends.hdf5_01.COLLECTION_COUNT = 5
    hangar.backends.hdf5_01.COLLECTION_SIZE = 20
    hangar.backends.numpy_10.COLLECTION_SIZE = 50
    hangar.backends.lmdb_30.LMDB_SETTINGS['map_size'] = 1_000_000

    old_map_size = hangar.constants.LMDB_SETTINGS['map_size']
    hangar.constants.LMDB_SETTINGS['map_size'] = 2_000_000
    hangar.txnctx.TxnRegisterSingleton._instances = {}

    pth = tmp_path_factory.mktemp('classrepo')
    repo_obj = Repository(path=str(pth), exists=False)
    repo_obj.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    yield repo_obj
    hangar.constants.LMDB_SETTINGS['map_size'] = old_map_size
    hangar.backends.hdf5_00.COLLECTION_COUNT = old00_count
    hangar.backends.hdf5_00.COLLECTION_SIZE = old00_size
    hangar.backends.hdf5_01.COLLECTION_COUNT = old01_count
    hangar.backends.hdf5_01.COLLECTION_SIZE = old01_size
    hangar.backends.numpy_10.COLLECTION_SIZE = old10_size
    hangar.backends.lmdb_30.LMDB_SETTINGS = old30_lmdb_settings
    repo_obj._env._close_environments()


@pytest.fixture()
def managed_tmpdir(monkeypatch, tmp_path):
    monkeypatch.setitem(hangar.constants.LMDB_SETTINGS, 'map_size', 2_000_000)
    monkeypatch.setitem(hangar.backends.lmdb_30.LMDB_SETTINGS, 'map_size', 1_000_000)
    monkeypatch.setattr(hangar.backends.hdf5_00, 'COLLECTION_COUNT', 5)
    monkeypatch.setattr(hangar.backends.hdf5_00, 'COLLECTION_SIZE', 20)
    monkeypatch.setattr(hangar.backends.hdf5_01, 'COLLECTION_COUNT', 5)
    monkeypatch.setattr(hangar.backends.hdf5_01, 'COLLECTION_SIZE', 20)
    monkeypatch.setattr(hangar.backends.numpy_10, 'COLLECTION_SIZE', 50)
    monkeypatch.setitem(hangar.backends.lmdb_30.LMDB_SETTINGS, 'map_size', 1_000_000)
    hangar.txnctx.TxnRegisterSingleton._instances = {}
    yield tmp_path
    shutil.rmtree(tmp_path)


@pytest.fixture()
def repo(managed_tmpdir) -> Repository:
    repo_obj = Repository(path=managed_tmpdir, exists=False)
    repo_obj.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    yield repo_obj
    repo_obj._env._close_environments()


@pytest.fixture()
def aset_samples_initialized_repo(repo) -> Repository:
    co = repo.checkout(write=True)
    co.add_ndarray_column(name='writtenaset', shape=(5, 7), dtype=np.float64)
    co.commit('this is a commit message')
    co.close()
    yield repo


@pytest.fixture()
def aset_subsamples_initialized_repo(repo) -> Repository:
    co = repo.checkout(write=True)
    co.add_ndarray_column(
        name='writtenaset', shape=(5, 7), dtype=np.float64, contains_subsamples=True)
    co.commit('this is a commit message')
    co.close()
    yield repo


@pytest.fixture(params=fixed_shape_backend_params)
def repo_20_filled_samples(request, aset_samples_initialized_repo, array5by7) -> Repository:
    co = aset_samples_initialized_repo.checkout(write=True)
    second_aset = co.add_ndarray_column('second_aset', prototype=array5by7, backend=request.param)
    first_aset = co.columns['writtenaset']
    for i in range(0, 20):
        array5by7[:] = i
        first_aset[str(i)] = array5by7
        array5by7[:] = -i
        second_aset[str(i)] = array5by7
    co.commit('20 samples')
    co.close()
    yield aset_samples_initialized_repo


@pytest.fixture(params=fixed_shape_backend_params)
def repo_300_filled_samples(request, aset_samples_initialized_repo, array5by7) -> Repository:
    co = aset_samples_initialized_repo.checkout(write=True)
    aset = co.add_ndarray_column('aset', prototype=array5by7, backend=request.param)
    with aset:
        for i in range(300):
            array5by7[:] = i
            aset[i] = array5by7
    co.commit('1000 samples')
    co.close()
    yield aset_samples_initialized_repo


@pytest.fixture()
def repo_20_filled_samples2(repo) -> Repository:
    # for diff testing
    dummyData = np.arange(50).astype(np.int64)
    co1 = repo.checkout(write=True, branch='master')
    co1.add_ndarray_column(name='dummy', prototype=dummyData)
    for idx in range(10):
        dummyData[:] = idx
        co1.columns['dummy'][idx] = dummyData
    co1.commit('first commit adding dummy data and hello meta')
    co1.close()
    return repo


@pytest.fixture(params=variable_shape_backend_params)
def aset_samples_var_shape_initialized_repo(request, repo) -> Repository:
    co = repo.checkout(write=True)
    co.add_ndarray_column(
        name='writtenaset', shape=(10, 10), dtype=np.float64, variable_shape=True, backend=request.param)
    co.commit('this is a commit message')
    co.close()
    yield repo


@pytest.fixture()
def aset_samples_initialized_w_checkout(aset_samples_initialized_repo) -> WriterCheckout:
    co = aset_samples_initialized_repo.checkout(write=True)
    yield co
    co.close()


@pytest.fixture()
def array5by7():
    return np.random.random((5, 7))


@pytest.fixture()
def randomsizedarray():
    a = random.randint(2, 8)
    b = random.randint(2, 8)
    return np.random.random((a, b))


@pytest.fixture(params=fixed_shape_backend_params)
def two_commit_filled_samples_repo(request, repo, array5by7) -> Repository:
    co = repo.checkout(write=True)
    co.add_ndarray_column(
        name='writtenaset', shape=(5, 7), dtype=np.float32, backend=request.param)
    for cIdx in range(2):
        if cIdx != 0:
            co = repo.checkout(write=True)

        with co.columns['writtenaset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range((cIdx + 1) * 5):
                arr = np.random.randn(*array5by7.shape).astype(np.float32) * 100
                d[str(sIdx)] = arr
        co.commit(f'commit number: {cIdx}')
        co.close()
    yield repo


@pytest.fixture()
def repo_1_br_no_conf(repo) -> Repository:

    dummyData = np.arange(50)
    co1 = repo.checkout(write=True, branch='master')
    co1.add_ndarray_column(name='dummy', prototype=dummyData)
    for idx in range(10):
        dummyData[:] = idx
        co1.columns['dummy'][str(idx)] = dummyData
    co1.commit('first commit adding dummy data')
    co1.close()

    repo.create_branch('testbranch')
    co2 = repo.checkout(write=True, branch='testbranch')
    for idx in range(10, 20):
        dummyData[:] = idx
        co2.columns['dummy'][str(idx)] = dummyData
        co2.columns['dummy'][idx] = dummyData
    co2.commit('first commit on test branch adding non-conflict data')
    co2.close()
    return repo


@pytest.fixture()
def repo_2_br_no_conf(repo_1_br_no_conf) -> Repository:

    dummyData = np.arange(50)
    repo = repo_1_br_no_conf
    co1 = repo.checkout(write=True, branch='master')
    for idx in range(20, 30):
        dummyData[:] = idx
        co1.columns['dummy'][str(idx)] = dummyData
        co1.columns['dummy'][idx] = dummyData
    co1.commit('second commit on master adding non-conflict data')
    co1.close()
    return repo


@pytest.fixture()
def server_instance(managed_tmpdir, worker_id, monkeypatch):
    from hangar.remote.server import serve

    address = f'localhost:{randint(50000, 59999)}'
    base_tmpdir = pjoin(managed_tmpdir, f'{worker_id[-1]}')
    mkdir(base_tmpdir)

    server, hangserver, _ = serve(base_tmpdir, overwrite=True, channel_address=address)
    monkeypatch.setitem(hangserver.CFG['SERVER_GRPC'], 'max_concurrent_rpcs', '10')
    monkeypatch.setitem(hangserver.CFG['SERVER_GRPC'], 'max_thread_pool_workers', '5')
    server.start()
    yield address

    hangserver.close()
    server.stop(0.05)
    time.sleep(0.1)
    if platform.system() == 'Windows':
        time.sleep(0.1)


@pytest.fixture()
def written_two_cmt_server_repo(server_instance, two_commit_filled_samples_repo) -> tuple:
    time.sleep(0.1)  # wait for ready
    two_commit_filled_samples_repo.remote.add('origin', server_instance)
    success = two_commit_filled_samples_repo.remote.push('origin', 'master')
    assert success == 'master'
    yield (server_instance, two_commit_filled_samples_repo)
